"""
problog.kbest - K-Best inference using MaxSat
---------------------------------------------

Anytime evaluation using best proofs.

..
    Part of the ProbLog distribution.

    Copyright 2015 KU Leuven, DTAI Research Group

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

from __future__ import print_function


from .core import transform
from .formula import LogicDAG
from .constraint import TrueConstraint, ClauseConstraint

from .cnf_formula import CNF, clarks_completion
from .maxsat import get_solver
from .evaluator import Evaluator, Evaluatable
from .logic import Term

from copy import deepcopy

import warnings
import logging

from functools import total_ordering


class KBestFormula(CNF, Evaluatable):

    transform_preference = 40

    def __init__(self, **kwargs):
        CNF.__init__(self)
        Evaluatable.__init__(self)

    def _create_evaluator(self, semiring, weights, **kwargs):
        return KBestEvaluator(self, semiring, weights, **kwargs)

    @classmethod
    def is_available(cls):
        """Checks whether the SDD library is available."""
        return True


transform(LogicDAG, KBestFormula, clarks_completion)


class KBestEvaluator(Evaluator):

    def __init__(self, formula, semiring, weights=None, lower_only=False,
                 verbose=None, convergence=1e-9, explain=None, **kwargs):
        Evaluator.__init__(self, formula, semiring, weights, **kwargs)

        self.sdd_manager = None
        # self.sdd_manager = SDDManager()
        # for l in range(1, formula.atomcount + 1):
        #     self.sdd_manager.add_variable(l)

        self._z = None
        self._weights = None
        self._given_weights = weights
        self._verbose = verbose
        self._lower_only = lower_only
        self._explain = explain
        if explain is not None:
            self._lower_only = True

        self._reverse_names = {index: name for name, index in self.formula.get_names()}

        self._convergence = convergence

    def initialize(self):
        raise NotImplementedError('Evaluator.initialize() is an abstract method.')

    def propagate(self):
        self._weights = self.formula.extract_weights(self.semiring, self._given_weights)
        self._z = self.semiring.one()
        # self._z = self.sdd_manager.wmc_true(self._weights, self.semiring)

    def evaluate(self, index):
        """Compute the value of the given node."""

        name = [n for n, i, l in self.formula.labeled() if i == index]
        if name:
            name = name[0]
        else:
            name = index

        logger = logging.getLogger('problog')
        logger.debug('evaluating query %s' % name)

        if index is None:
            if self._explain is not None:
                self._explain.append('%s :- fail.' % name)
            return 0.0
        elif index == 0:
            if self._explain is not None:
                self._explain.append('%s :- true.' % name)
            return 1.0
        else:
            lb = Border(self.formula, self.sdd_manager, self.semiring, index, 'lower')
            ub = Border(self.formula, self.sdd_manager, self.semiring, -index, 'upper')

            k = 0
            # Select the border with most improvement
            if self._lower_only:
                nborder = lb
            else:
                nborder = max(lb, ub)

            try:
                while not nborder.is_complete():
                    solution = nborder.update()
                    logger.debug('  update: %s %s < p < %s ' %
                                 (nborder.name, lb.value, 1.0 - ub.value))
                    if self._explain is not None and solution is not None:
                        solution_names = []

                        probability = nborder.improvement
                        for s in solution:
                            n = self._reverse_names.get(abs(s), Term('choice_%s' % (abs(s))))
                            if s < 0:
                                solution_names.append(-n)
                            else:
                                solution_names.append(n)
                        proof = ', '.join(map(str, solution_names))
                        self._explain.append('%s :- %s.  %% P=%.8g' % (name, proof, probability))

                    if solution is not None:
                        k += 1

                    if nborder.is_complete():
                        if nborder == lb:
                            if self._explain is not None:
                                if k == 0:
                                    self._explain.append('%s :- fail.' % name)
                                self._explain.append('')
                            return lb.value
                        else:
                            return 1.0 - ub.value

                    if ub.value + lb.value > 1.0 - self._convergence:
                        logger.debug('  convergence reached')
                        return lb.value, 1.0 - ub.value
                    if self._lower_only:
                        nborder = lb
                    else:
                        nborder = max(lb, ub)
            except KeyboardInterrupt:
                pass
            except SystemError:
                pass

            return lb.value, 1.0 - ub.value

    def evaluate_evidence(self):
        raise NotImplementedError('Evaluator.evaluate_evidence is an abstract method.')

    def add_evidence(self, node):
        """Add evidence"""
        warnings.warn('Evidence is not supported by this evaluation method and will be ignored.')

    def has_evidence(self):
        return self.__evidence != []

    def clear_evidence(self):
        self.__evidence = []

    def evidence(self):
        return iter(self.__evidence)


@total_ordering
class Border(object):

    def __init__(self, cnf, manager, semiring, query, name, smart_constraints=False):
        self.wcnf = deepcopy(cnf)
        self.wcnf.add_constraint(TrueConstraint(query), True)

        self.name = name

        self.manager = manager
        self.semiring = semiring

        self.weights = self.wcnf.extract_weights(self.semiring)
        # self.compiled = self.manager.false()

        self.value = 0.0
        self.improvement = 1.0

        self.smart_constraints = smart_constraints

    def update(self):
        solver = get_solver()

        solution = solver.evaluate(self.wcnf, partial=True, smart_constraints=True)

        if solution is None:
            self.improvement = None
            return None
        else:
            # m = self.manager

            solution = self.wcnf.from_partial(solution)

            probability = self.semiring.one()
            for s in solution:
                wp, wn = self.weights[abs(s)]
                if s < 0:
                    probability = self.semiring.times(probability, wn)
                else:
                    probability = self.semiring.times(probability, wp)
            probability = self.semiring.result(probability)

            constraint = ClauseConstraint(list(map(lambda x: -x, solution)))
            self.wcnf.add_constraint(constraint, True)
            # literals = list(map(m.literal, solution))

            # proof_sdd = m.conjoin(*literals)
            # sdd_query_new = m.disjoin(self.compiled, proof_sdd)
            # m.deref(proof_sdd, self.compiled)
            # self.compiled = sdd_query_new
            #
            # pquery = m.wmc(self.compiled, self.weights, self.semiring)
            # ptrue = m.wmc_true(self.weights, self.semiring)
            # res = self.semiring.normalize(pquery, ptrue)

            # value = self.semiring.result(res)

            # self.improvement = value - self.value
            # self.value = value

            self.improvement = probability
            self.value = self.value + probability

            assert abs(self.improvement - probability) < 1e-8

            return solution

    def is_complete(self):
        return self.improvement is None

    def __lt__(self, other):
        if self.improvement is None:
            return True
        elif other.improvement is None:
            return False
        else:
            return self.improvement < other.improvement

    def __eq__(self, other):
        return self.improvement == other.improvement





#! /usr/bin/env python

"""
ProbLog command-line interface.

Copyright 2015 KU Leuven, DTAI Research Group

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from __future__ import print_function

import sys
import traceback

from problog.program import PrologFile
from problog.constraint import TrueConstraint
from problog.formula import LogicFormula, LogicDAG
from problog.cnf_formula import CNF
from problog.maxsat import get_solver, get_available_solvers
from problog.errors import process_error
from problog import get_evaluatable
from problog.evaluator import Semiring, OperationNotSupported, SemiringProbability
from problog.logic import Term
from problog.util import init_logger, Timer


def main(argv):
    args = argparser().parse_args(argv)

    if args.use_semiring:
        return mpe_semiring(args)
    else:
        return mpe_maxsat(args)


def mpe_semiring(args):
    inputfile = args.inputfile

    init_logger(args.verbose)

    if args.web:
        result_handler = print_result_json
    else:
        result_handler = print_result

    if args.output is not None:
        outf = open(args.output, 'w')
    else:
        outf = sys.stdout

    with Timer("Total"):
        try:
            pl = PrologFile(inputfile)

            semiring = SemiringMPEState()
            kc_class = get_evaluatable(name='sdd')        # TODO should pass semiring
                                                # -> no compilation required for MPE
                                                # BUT semiring does not support negation
                                                # so formula should be NNF (only negation of facts)
            lf = LogicFormula.create_from(pl, label_all=True)

            if lf.evidence():
                # Query = evidence + constraints
                qn = lf.add_and([y for x, y in lf.evidence()])
                lf.clear_evidence()

                if lf.queries():
                    print ('%% WARNING: ignoring queries in file', file=sys.stderr)
                lf.clear_queries()

                query_name = Term('query')
                lf.add_query(query_name, qn)
                kc = kc_class.create_from(lf)

                with open('/tmp/x.dot', 'w') as f:
                    print (kc.to_formula().to_dot(), file=f)

                results = kc.evaluate(semiring=semiring)
                prob, facts = results[query_name]
            else:
                prob, facts = 1.0, []

            result_handler((True, (prob, facts)), outf)

        except Exception as err:
            trace = traceback.format_exc()
            err.trace = trace
            result_handler((False, err), outf)


def mpe_maxsat(args):
    inputfile = args.inputfile

    if args.web:
        result_handler = print_result_json
    else:
        result_handler = print_result

    if args.output is not None:
        outf = open(args.output, 'w')
    else:
        outf = sys.stdout

    logger = init_logger(args.verbose)

    with Timer("Total"):
        try:
            pl = PrologFile(inputfile)

            dag = LogicDAG.createFrom(pl, avoid_name_clash=True, label_all=True, labels=[('output', 1)])

            if dag.queries():
                print('%% WARNING: ignoring queries in file', file=sys.stderr)
                dag.clear_queries()

            logger.info('Ground program size: %s' % len(dag))

            cnf = CNF.createFrom(dag)

            for qn, qi in cnf.evidence():
                if not cnf.is_true(qi):
                    cnf.add_constraint(TrueConstraint(qi))

            queries = list(cnf.labeled())

            logger.info('CNF size: %s' % cnf.clausecount)

            if not cnf.is_trivial():
                solver = get_solver(args.solver)

                with Timer('Solving'):
                    result = frozenset(solver.evaluate(cnf))
                weights = cnf.extract_weights(SemiringProbability())
                output_facts = None
                prob = 1.0
                if result is not None:
                    output_facts = []

                    if queries:
                        for qn, qi, ql in queries:
                            if qi in result:
                                output_facts.append(qn)
                            elif -qi in result:
                                output_facts.append(-qn)
                    for i, n, t in dag:
                        if t == 'atom':
                            if i in result:
                                if not queries:
                                    output_facts.append(n.name)
                                prob *= weights[i][0]
                            elif -i in result:
                                if not queries:
                                    output_facts.append(-n.name)
                                prob *= weights[i][1]
            else:
                prob = 1.0
                output_facts = []

            result_handler((True, (prob, output_facts)), outf)
        except Exception as err:
            trace = traceback.format_exc()
            err.trace = trace
            result_handler((False, err), outf)

        if args.output is not None:
            outf.close()


def print_result(result, output=sys.stdout):
    success, result = result
    if success:
        prob, facts = result
        if facts is None:
            print ('%% The model is not satisfiable.', file=output)
        else:
            for atom in facts:
                print(atom, file=output)
            if prob is not None:
                print ('%% Probability: %.10g' % prob)
        return 0
    else:
        print(process_error(result), file=output)
        return 1


def print_result_json(d, output):
    """Pretty print result.

    :param d: result from run_problog
    :param output: output file
    :param precision:
    :return:
    """
    import json
    result = {}
    success, d = d
    if success:
        prob, facts = d
        result['SUCCESS'] = True
        if facts is not None:
            result['atoms'] = list(map(lambda n: (str(-n), False) if n.is_negated() else (str(n), True), facts))
        if prob is not None:
            result['prob'] = round(prob, 10)
    else:
        result['SUCCESS'] = False
        result['err'] = process_error(d)
        result['original'] = str(d)
    print (json.dumps(result), file=output)
    return 0


def reduce_formula(formula, facts):
    # Assume formula is cycle free.

    values = [None] * len(formula)
    for f in facts:
        if f > 0:
            values[f - 1] = True
        else:
            values[-f - 1] = False

    from collections import deque
    nodes = deque(range(1, len(formula) + 1))
    while nodes:
        index = nodes.popleft()
        value = values[index - 1]
        if value is None:
            node = formula.get_node(index)
            nodetype = type(node).__name__
            if nodetype == 'atom':
                pass  # Really shouldn't happen
            else:
                children = [values[c] if c > 0 else not values[-c] for c in node.children]
                if nodetype == 'disj':
                    if True in children:
                        values[index - 1] = True
                    elif None in children:
                        # not ready yet, push it back on the queue
                        nodes.append(index)
                    else:
                        values[index - 1] = False
                else:
                    if False in children:
                        values[index - 1] = False
                    elif None in children:
                        nodes.append(index)
                    else:
                        values[index - 1] = True
    return values


class SemiringMPEState(Semiring):

    def __init__(self):
        Semiring.__init__(self)

    def zero(self):
        return 0.0, set()

    def one(self):
        return 1.0, set()

    def plus(self, a, b):
        if a[0] > b[0]:
            return a
        elif a[0] < b[0]:
            return b
        else:
            return a[0], a[1]  # | b[1]   # doesn't matter?

    def times(self, a, b):
        return a[0] * b[0], a[1] | b[1]

    def pos_value(self, a, key=None):
        return float(a), {key}

    def neg_value(self, a, key=None):
        return 1.0 - float(a), {-key}

    def is_nsp(self):
        return True

    def result(self, a, formula=None):
        return a

    def ad_complement(self, ws, key=None):
        s = sum([x[0] for x in ws])
        return 1.0 - s, {key}


def argparser():
    import argparse

    parser = argparse.ArgumentParser()
    parser.add_argument('inputfile')
    parser.add_argument('--solver', choices=get_available_solvers(),
                        default=None, help="MaxSAT solver to use")
    parser.add_argument('--full', dest='output_all', action='store_true',
                        help='Also show false atoms.')
    parser.add_argument('-o', '--output', type=str, default=None,
                        help='Write output to given file (default: write to stdout)')
    parser.add_argument('--web', action='store_true', help=argparse.SUPPRESS)
    parser.add_argument('--use-maxsat', action='store_true', help=argparse.SUPPRESS)
    parser.add_argument('--use-semiring', action='store_true', help=argparse.SUPPRESS)
    parser.add_argument('-v', '--verbose', action='count', help='Increase verbosity')
    return parser


if __name__ == '__main__':
    main(sys.argv[1:])



"""
problog.util - Useful utilities
-------------------------------

Provides useful utilities functions and classes.

..
    Part of the ProbLog distribution.

    Copyright 2015 KU Leuven, DTAI Research Group

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""
from __future__ import print_function

import logging
import time

import signal
import sys
import os
import subprocess
import distutils.spawn
import tempfile
import imp
import collections


class ProbLogLogFormatter(logging.Formatter):

    def __init__(self):
        logging.Formatter.__init__(self)

    def format(self, message):
        msg = str(message.msg) % message.args
        lines = msg.split('\n')
        if message.levelno < 10:
            linestart = '[LVL%s] ' % message.levelno
        else:
            linestart = '[%s] ' % message.levelname
        return linestart + ('\n' + linestart).join(lines)


def init_logger(verbose=None, name='problog', out=None):
    """Initialize default logger.

    :param verbose: verbosity level (0: WARNING, 1: INFO, 2: DEBUG)
    :type verbose: int
    :param name: name of the logger (default: problog)
    :type name: str
    :return: result of ``logging.getLogger(name)``
    :rtype: logging.Logger
    """
    if out is None:
        out = sys.stdout

    logger = logging.getLogger(name)
    ch = logging.StreamHandler(out)
    # formatter = logging.Formatter('[%(levelname)s] %(message)s')
    formatter = ProbLogLogFormatter()
    ch.setFormatter(formatter)
    logger.addHandler(ch)
    if not verbose:
        logger.setLevel(logging.WARNING)
    elif verbose == 1:
        logger.setLevel(logging.INFO)
        logger.info('Output level: INFO')
    elif verbose == 2:
        logger.setLevel(logging.DEBUG)
        logger.debug('Output level: DEBUG')
    else:
        level = max(1, 12 - verbose)   # between 9 and 1
        logger.setLevel(level)
        logger.log(level, 'Output level: %s' % level)
    return logger


class Timer(object):
    """Report timing information for a block of code.
    To be used as a ``with`` block.

    :param msg: message to print
    :type msg: str
    :param output: file object to write to (default: write to logger ``problog``)
    :type output: file
    """

    def __init__(self, msg, output=None, logger='problog'):
        self.message = msg
        self.start_time = None
        self.output = output
        self.logger = logger

    def __enter__(self):
        self.start_time = time.time()

    # noinspection PyUnusedLocal
    def __exit__(self, *args):
        if self.output is None:
            logger = logging.getLogger(self.logger)
            logger.info('%s: %.4fs' % (self.message, time.time() - self.start_time))
        else:
            print ('%s: %.4fs' % (self.message, time.time() - self.start_time), file=self.output)


# noinspection PyUnusedLocal
def _raise_timeout(*args):
    """Raise global timeout exception (used by global timer)

    :param args: signal information (ignored)
    :raise KeyboardInterrupt:
    """
    raise KeyboardInterrupt('Timeout')   # Global exception on all threads


def start_timer(timeout=0):
    """Start a global timeout timer.

    :param timeout: timeout in seconds
    :type timeout: int
    """
    signal.signal(signal.SIGALRM, _raise_timeout)
    signal.alarm(timeout)


def stop_timer():
    """Stop the global timeout timer."""
    signal.alarm(0)


def subprocess_check_output(*popenargs, **kwargs):
    """Wrapper for subprocess.check_output that recursively kills subprocesses when Python is \
    interrupted.

    Additionally expands executable name to full path.

    :param popenargs: positional arguments of subprocess.call
    :param kwargs: keyword arguments of subprocess.call
    :return: result of subprocess.call
    """
    popenargs = _find_process(*popenargs)
    process = subprocess.Popen(stdout=subprocess.PIPE, *popenargs, **kwargs)
    try:
        output, unused_err = process.communicate()
        retcode = process.poll()
        if retcode:
            cmd = kwargs.get("args")
            if cmd is None:
                cmd = popenargs[0]
            raise subprocess.CalledProcessError(retcode, cmd, output=output)
        return output.decode()
    except KeyboardInterrupt:
        kill_proc_tree(process)
        raise
    except SystemExit:
        kill_proc_tree(process)
        raise


def subprocess_check_call(*popenargs, **kwargs):
    """Wrapper for subprocess.check_call that recursively kills subprocesses when Python is \
    interrupted.

    Additionally expands executable name to full path.

    :param popenargs: positional arguments of subprocess.call
    :param kwargs: keyword arguments of subprocess.call
    :return: result of subprocess.call
    """
    retcode = subprocess_call(*popenargs, **kwargs)
    if retcode:
        cmd = kwargs.get("args")
        if cmd is None:
            cmd = popenargs[0]
        raise subprocess.CalledProcessError(retcode, cmd)
    return 0


def subprocess_call(*popenargs, **kwargs):
    """Wrapper for subprocess.call that recursively kills subprocesses when Python is interrupted.

    Additionally expands executable name to full path.

    :param popenargs: positional arguments of subprocess.call
    :param kwargs: keyword arguments of subprocess.call
    :return: result of subprocess.call
    """
    process = None
    try:
        popenargs = _find_process(*popenargs)
        process = subprocess.Popen(*popenargs, **kwargs)
        return process.wait()
    except KeyboardInterrupt:
        kill_proc_tree(process)
        raise
    except SystemExit:
        kill_proc_tree(process)
        raise


def _find_process(cmd, *rest):
    fullname = distutils.spawn.find_executable(cmd[0])
    if fullname is not None:
        return ([fullname] + cmd[1:],) + rest
    else:
        return (cmd,) + rest


def kill_proc_tree(process, including_parent=True):
    """Recursively kill a subprocess. Useful when the subprocess is a script.
    Requires psutil but silently fails when it is not present.

    :param process: process
    :type process: subprocess.Popen
    :param including_parent: also kill process itself (default: True)
    :type including_parent: bool
    """
    try:
        # noinspection PyPackageRequirements
        import psutil
        pid = process.pid
        parent = psutil.Process(pid)
        children = parent.children(recursive=True)
        for child in children:
            child.kill()
        psutil.wait_procs(children, timeout=5)
        if including_parent:
            parent.kill()
            parent.wait(5)
    except ImportError:
        # noinspection PyUnusedLocal
        psutil = None
        process.kill()


class OrderedSet(collections.MutableSet):
    """Provides an ordered version of a set which keeps elements in the order they are added.

    :param iterable: add elements from this iterable (default: None)
    :type iterable: Sequence
    """

    def __init__(self, iterable=None):
        self.end = end = []
        end += [None, end, end]         # sentinel node for doubly linked list
        self.map = {}                   # key --> [key, prev, next]
        if iterable is not None:
            # noinspection PyMethodFirstArgAssignment
            self |= iterable

    def __len__(self):
        return len(self.map)

    def __contains__(self, key):
        return key in self.map

    def add(self, key):
        """Add element.

        :param key: element to add
        """
        if key not in self.map:
            end = self.end
            curr = end[1]
            curr[2] = end[1] = self.map[key] = [key, curr, end]

    def discard(self, key):
        """Discard element.

        :param key: element to remove
        """
        if key in self.map:
            key, prv, nxt = self.map.pop(key)
            prv[2] = nxt
            nxt[1] = prv

    def __iter__(self):
        end = self.end
        curr = end[2]
        while curr is not end:
            yield curr[0]
            curr = curr[2]

    def __reversed__(self):
        end = self.end
        curr = end[1]
        while curr is not end:
            yield curr[0]
            curr = curr[1]

    def pop(self, last=True):
        """Remove and return first or last element.

        :param last: remove last element
        :return: last element
        """
        if not self:
            raise KeyError('set is empty')
        key = self.end[1][0] if last else self.end[2][0]
        self.discard(key)
        return key

    def __repr__(self):
        if not self:
            return '%s()' % (self.__class__.__name__,)
        return '%s(%r)' % (self.__class__.__name__, list(self))

    def __eq__(self, other):
        if other is None:
            return False
        elif isinstance(other, OrderedSet):
            return len(self) == len(other) and list(self) == list(other)
        return set(self) == set(other)


def mktempfile(suffix=''):
    """Create a temporary file with the given name suffix.

    :param suffix: extension of the file
    :type suffix: str
    :return: name of the temporary file
    """
    fd, filename = tempfile.mkstemp(suffix)
    os.close(fd)
    return filename


def load_module(filename):
    """Load a Python module from a filename or qualified module name.

    If filename ends with ``.py``, the module is loaded from the given file.
    Otherwise it is taken to be a module name reachable from the path.

    Example:

    .. code-block: python

       pb_util = load_module('problog.util')
       pb_util = load_module('problog/util.py')

    :param filename: location of the module
    :type filename: str
    :return: loaded module
    :rtype: module
    """
    if filename.endswith('.py'):
        filename = os.path.abspath(os.path.join(os.path.dirname(__file__), filename))
        (path, name) = os.path.split(filename)
        (name, ext) = os.path.splitext(name)
        (extfile, filename, data) = imp.find_module(name, [path])
        return imp.load_module(name, extfile, filename, data)
    else:
        mod = __import__(filename)
        components = filename.split('.')
        for c in components[1:]:
            mod = getattr(mod, c)
        return mod


def format_value(data, precision=8):
    """Pretty print a given value.

    :param data: data to format
    :param precision: max. number of digits
    :type precision: int
    :return: pretty printed result
    :rtype: str
    """
    if isinstance(data, float):
        data = ('%.' + str(precision) + 'g') % data
    else:
        data = str(data)
    return ('{:<' + str(precision + 2) + '}').format(data)


def format_tuple(data, precision=8, columnsep='\t'):
    """Pretty print a given tuple (or single value).

    :param data: data to format
    :param precision: max. number of digits
    :type precision: int
    :param columnsep: column separator
    :type columnsep: str
    :return: pretty printed result
    :rtype: str
    """

    if isinstance(data, str):
        # is a string -> return string itself
        return data
    elif isinstance(data, collections.Sequence):
        values = list(map(lambda v: format_value(v, precision=precision), data))
        if len(values) == 2 and values[0] == values[1]:
            values = [values[0]]
        return columnsep.join(values)
    else:
        return format_value(data, precision=precision)


def format_dictionary(data, precision=8, keysep=':', columnsep='\t'):
    """Pretty print a given dictionary.

    :param data: data to format
    :type data: dict
    :param precision: max. number of digits
    :type precision: int
    :param keysep: separator between key and value (default: ``;``)
    :type keysep: str
    :param columnsep: column separator (default: ``tab``)
    :type columnsep: str
    :return: pretty printed result
    :rtype: str
    """
    if not data:
        return ""  # no queries
    else:
        s = []
        # Determine maximum length of key
        l = max(len(str(k)) for k in data)
        fmt = ('{:>' + str(l) + '}{}{}{}')
        for it in sorted([(str(k), v) for k, v in data.items()]):
            val = format_tuple(it[1], precision=precision, columnsep=columnsep)
            s.append(fmt.format(it[0], keysep, columnsep, val))
        return '\n'.join(s)


class UHeap(object):
    """Updatable heap.

    Each element is represented as a pair (key, item).
    The operation ``pop()`` always returns the item with the smallest key.
    The operation ``push(item)`` either adds item (returns True) or updates its key (return False)
    A function for computing an item's key can be passed.

    :param key: function for computing the sort key of an item
    """

    def __init__(self, key=None):
        self._heap = []
        self._index = {}
        self._key = key

    def _compute_key(self, item):
        if self._key is None:
            return item
        else:
            return self._key(item)

    def push(self, item):
        """Add the item or update it's key in case it already exists.

        :param item: item to add
        :return: True is item was not in the collection
        """
        # Compute the item's key
        key = self._compute_key(item)
        # Check if element is already there.
        index = self._index.get(item)
        if index is None:
            # It is not: normal add
            self._heap.append((key, item))
            index = len(self._heap) - 1
            self._index[item] = index
            self._swim_up(index)
            is_new = True
        else:
            # Element is there
            oldkey, item = self._heap[index]
            if oldkey == key:
                # Keys are the same, nothing to do.
                pass
            else:
                # Replace value
                self._heap[index] = (key, item)
                # Compare value with parent and swim up or sink down accordingly.
                parent = self._parent(index)
                if parent is not None and key < self._heap[parent][0]:
                    self._swim_up(index)
                else:
                    self._sink_down(index)
            is_new = False
        return is_new

    def pop(self):
        """Removes and returns the element with the smallest key.

        :return: item with the smallest key
        """
        return self.pop_with_key()[1]

    def pop_with_key(self):
        """Removes and returns the smallest element and its key.

        :return: smallest element (key, element)
        """
        assert bool(self)
        # Get top element
        key, item = self._heap[0]
        # Swap top and last
        self._swap(0, len(self._heap) - 1)
        # Remove last element (former top)
        del self._index[item]
        self._heap.pop(-1)
        # Sink down new top element
        if self:
            self._sink_down(0)
        return key, item

    def peek(self):
        """Returns the element with the smallest key without removing it.

        :return: item with the smallest key
        """
        assert bool(self)
        # Get top element
        key, item = self._heap[0]
        return item

    def _parent(self, index):
        if index == 0:
            return None
        else:
            return (index - 1) // 2

    def _children(self, index):
        return (2 * index) + 1, (2 * index) + 2

    def _swim_up(self, index):
        p = self._parent(index)
        if p is not None and self._heap[p][0] > self._heap[index][0]:
            self._swap(p, index)
            self._swim_up(p)

    def _sink_down(self, index):
        c1, c2 = self._children(index)
        k1 = None
        k2 = None
        if c1 < len(self._heap):
            k1 = self._heap[c1][0]
            if c2 < len(self._heap):
                k2 = self._heap[c2][0]
        k = self._heap[index][0]
        if k1 is not None and k > k1:
            if k2 is not None and k1 > k2:
                self._swap(index, c2)
                self._sink_down(c2)
            else:
                self._swap(index, c1)
                self._sink_down(c1)
        elif k2 is not None and k > k2:
            self._swap(index, c2)
            self._sink_down(c2)

    def _swap(self, index1, index2):
        key1, item1 = self._heap[index1]
        key2, item2 = self._heap[index2]
        # Update index
        self._index[item1] = index2
        self._index[item2] = index1
        # Update heap
        self._heap[index1] = key2, item2
        self._heap[index2] = key1, item1

    def __len__(self):
        return len(self._heap)


class BitVector(object):

    def __init__(self):
        self.binsize_bits = 5
        self.binsize = 1 << self.binsize_bits
        # self.blocks = array.array('L')
        self.blocks = []
        self.blocks_size = []

    def add(self, index):
        # b = index // self.binsize
        # i = index % self.binsize
        mask = ((1 << self.binsize_bits) - 1)
        b = index >> self.binsize_bits
        i = index & mask
        n = len(self.blocks)
        if n <= b:
            self.blocks.extend([0] * (b - n + 1))
        self.blocks[b] |= (1 << i)

    def __contains__(self, index):
        mask = ((1 << self.binsize_bits) - 1)
        b = index >> self.binsize_bits
        i = index & mask
        n = len(self.blocks)
        if n <= b:
            return False
        else:
            return self.blocks[b] & (1 << i)

    def __iter__(self):
        o = 0
        for b, block in enumerate(self.blocks):
            if block != 0:
                for i in range(0, self.binsize):
                    if (1 << i) & block:
                        yield o + i
            o += self.binsize

    def __and__(self, other):
        result = BitVector()
        for a, b in zip(self.blocks, other.blocks):
            result.blocks.append(a & b)
        return result

    def __iand__(self, other):
        la = len(self.blocks)
        for i, b in enumerate(other.blocks[:la]):
            self.blocks[i] &= b
        return self

    def __or__(self, other):
        result = BitVector()
        for a, b in zip(self.blocks, other.blocks):
            result.blocks.append(a | b)

        la = len(self.blocks)
        lb = len(other.blocks)

        result.blocks.extend(self.blocks[lb:])
        result.blocks.extend(other.blocks[la:])
        return result

    def __ior__(self, other):
        la = len(self.blocks)
        for i, b in enumerate(other.blocks[:la]):
            self.blocks[i] |= b
        self.blocks.extend(other.blocks[la:])
        return self

    def __len__(self):
        n = 0
        for block in self.blocks:
            if block != 0:
                n += bin(block).count("1")
                # n += self._countbits(block)
        return n

    def __nonzero__(self):
        for b in self.blocks:
            if b:
                return True
        return False

    def __bool__(self):
        for b in self.blocks:
            if b:
                return True
        return False

    def _countbits(self, n):
        n = (n & 0x5555555555555555) + ((n & 0xAAAAAAAAAAAAAAAA) >> 1)
        n = (n & 0x3333333333333333) + ((n & 0xCCCCCCCCCCCCCCCC) >> 2)
        n = (n & 0x0F0F0F0F0F0F0F0F) + ((n & 0xF0F0F0F0F0F0F0F0) >> 4)
        n = (n & 0x00FF00FF00FF00FF) + ((n & 0xFF00FF00FF00FF00) >> 8)
        n = (n & 0x0000FFFF0000FFFF) + ((n & 0xFFFF0000FFFF0000) >> 16)
        n = (n & 0x00000000FFFFFFFF) + ((n & 0xFFFFFFFF00000000) >> 32)  # This last & isn't strictly necessary.
        return n


"""
problog.forward - Forward compilation and evaluation
----------------------------------------------------

Forward compilation using TP-operator.

..
    Part of the ProbLog distribution.

    Copyright 2015 KU Leuven, DTAI Research Group

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""
from __future__ import print_function
from .formula import LogicFormula, OrderedSet
from .dd_formula import DD
from .sdd_formula import SDD
from .bdd_formula import BDD
from .core import transform
from .evaluator import Evaluator, EvaluatableDSP

from .dd_formula import build_dd

import warnings
import time
import logging

import signal
from .core import transform_create_as

from .util import UHeap

import random

from collections import defaultdict


def timeout_handler(signum, frame):
    raise SystemError('Process timeout (Python) [%s]' % signum)


class ForwardInference(DD):
    def __init__(self, compile_timeout=None, **kwdargs):
        super(ForwardInference, self).__init__(auto_compact=False, **kwdargs)

        self._inodes_prev = None
        self._inodes_old = None
        self._inodes_neg = None
        self._facts = None
        self._atoms_in_rules = None
        self._completed = None

        self.timeout = compile_timeout

        self._update_listeners = []

        self._node_depths = None
        self.evidence_node = 0

    def register_update_listener(self, obj):
        self._update_listeners.append(obj)

    def _create_atom(self, identifier, probability, group, name=None, source=None):
        return self._atom(identifier, probability, group, name, source)

    def is_complete(self, node):
        node = abs(node)
        return self._completed[node - 1]

    def set_complete(self, node):
        self._completed[node - 1] = True

    def init_build(self):
        if self.evidence():
            self.evidence_node = self.add_and([n for q, n in self.evidence() if n is None or n != 0])
        self._facts = []  # list of facts
        self._atoms_in_rules = defaultdict(set)  # lookup all rules in which an atom is used
        self._completed = [False] * len(self)

        self._compute_node_depths()
        for index, node, nodetype in self:
            if self._node_depths[index - 1] is not None:
                # only include nodes that are reachable from a query or evidence
                if nodetype == 'atom':  # it's a fact
                    self._facts.append(index)
                    self.set_complete(index)
                else:  # it's a compound
                    for atom in node.children:
                        self._atoms_in_rules[abs(atom)].add(index)
        self.build_constraint_dd()
        self.inodes = [None] * len(self)
        self._inodes_prev = [None] * len(self)
        self._inodes_old = [None] * len(self)
        self._inodes_neg = [None] * len(self)
        self._compute_minmax_depths()

    def _propagate_complete(self, interrupted=False):
        if not interrupted:
            for i, c in enumerate(self._completed):
                if not c:
                    self._completed[i] = True
                    self.notify_node_completed(i + 1)
        else:
            updated_nodes = set([(i + 1) for i, c in enumerate(self._completed) if c])
            while updated_nodes:
                next_updates = set()
                # Find all heads that are affected
                affected_nodes = set()
                for node in updated_nodes:
                    for rule in self._atoms_in_rules[node]:
                        if not self.is_complete(rule):
                            affected_nodes.add(rule)
                for head in affected_nodes:
                    # head must be compound
                    node = self.get_node(head)
                    children = [self.is_complete(c) for c in node.children]
                    if False not in children:
                        self.is_complete(head)
                        self.notify_node_completed(head)
                        next_updates.add(head)
                updated_nodes = next_updates

    def _compute_node_depths(self):
        """Compute node depths in breadth-first manner."""
        self._node_depths = [None] * len(self)
        self._node_levels = []
        # Start with current nodes
        current_nodes = set(abs(n) for q, n, l in self.labeled() if self.is_probabilistic(n))
        if self.is_probabilistic(self.evidence_node):
            current_nodes.add(self.evidence_node)
        current_level = 0
        while current_nodes:
            self._node_levels.append(current_nodes)
            next_nodes = set()
            for index in current_nodes:
                self._node_depths[index - 1] = current_level
                node = self.get_node(index)
                nodetype = type(node).__name__
                if nodetype != 'atom':
                    for c in node.children:
                        if self.is_probabilistic(c):
                            if self._node_depths[abs(c) - 1] is None:
                                next_nodes.add(abs(c))
            current_nodes = next_nodes
            current_level += 1

    def _compute_minmax_depths(self):
        self._node_minmax = [None] * len(self)
        for level, nodes in reversed(list(enumerate(self._node_levels))):
            for index in nodes:
                # Get current node's minmax
                minmax = self._node_minmax[index - 1]
                if minmax is None:
                    minmax = level
                for rule in self._atoms_in_rules[index]:
                    rule_minmax = self._node_minmax[rule - 1]
                    if rule_minmax is None:
                        self._node_minmax[rule - 1] = minmax
                    else:
                        node = self.get_node(rule)
                        nodetype = type(node).__name__
                        if nodetype == 'conj':
                            rule_minmax = max(minmax, rule_minmax)
                        else:  # disj
                            rule_minmax = min(minmax, rule_minmax)
                        self._node_minmax[rule - 1] = rule_minmax

    def _update_minmax_depths(self, index, new_minmax=0):
        """Update the minmax depth data structure when the given node is completed.

        :param index:
        :return:
        """
        current_minmax = self._node_minmax[index - 1]
        self._node_minmax[index - 1] = new_minmax

        for parent in self._atoms_in_rules[index]:
            parent_minmax = self._node_minmax[parent - 1]

            if current_minmax == parent_minmax:
                # Current node is best child => we need to recompute
                parent_node = self.get_node(parent)
                parent_nodetype = type(parent_node).__name__
                parent_children_minmax = [self._node_minmax[c - 1]
                                          for c in parent_node.children
                                          if not self.is_complete(c)]
                if not parent_children_minmax:
                    # No incomplete children
                    self.set_complete(parent)
                    parent_minmax = 0
                elif parent_nodetype == 'conj':
                    parent_minmax == max(parent_children_minmax)
                else:
                    parent_minmax == min(parent_children_minmax)
                self._update_minmax_depths(parent, parent_minmax)

    def sort_nodes(self, nodes):
        return sorted(nodes, key=lambda i: self._node_depths[i - 1])

    def notify_node_updated(self, node, complete):
        for obj in self._update_listeners:
            obj.node_updated(self, node, complete)

    def notify_node_completed(self, node):
        for obj in self._update_listeners:
            obj.node_completed(self, node)

    def _heuristic_key_depth(self, node):
        # For OR: D(n) is min(D(c) for c in children)
        # For AND: D(n) is max(D(c) for c in children)
        return self._node_minmax[node - 1], self._node_depths[node - 1], random.random()

    def _heuristic_key(self, node):
        return self._heuristic_key_depth(node)

    def build_iteration(self, updated_nodes):
        to_recompute = UHeap(key=self._heuristic_key)
        for node in updated_nodes:
            for rule in self._atoms_in_rules[node]:
                to_recompute.push(rule)

        # nodes_to_recompute should be an updateable heap without duplicates
        while to_recompute:
            key, node = to_recompute.pop_with_key()
            # print ('recompute node:', node, len(to_recompute), key, self.get_node(node))
            if self.update_inode(node):  # The node has changed
                # Find rules that may be affected
                for rule in self._atoms_in_rules[node]:
                    to_recompute.push(rule)
                # Notify listeners that node was updated
                self.notify_node_updated(node, self.is_complete(node))
            elif self.is_complete(node):
                self.notify_node_completed(node)
                # if self.is_complete(node):
                #     self._update_minmax_depths(node)

    def build_iteration_levelwise(self, updated_nodes):
        while updated_nodes:
            next_updates = OrderedSet()
            # Find all heads that are affected
            affected_nodes = OrderedSet()
            for node in updated_nodes:
                for rule in self._atoms_in_rules[node]:
                    affected_nodes.add(rule)
            affected_nodes = self.sort_nodes(affected_nodes)
            # print (affected_nodes, [self._node_depths[i-1] for i in affected_nodes])
            for head in affected_nodes:
                if self.update_inode(head):
                    next_updates.add(head)
                    self.notify_node_updated(head, self.is_complete(head))
                elif self.is_complete(head):
                    self.notify_node_completed(head)
            updated_nodes = next_updates

    def build_stratum(self, updated_nodes):
        self.build_iteration(updated_nodes)
        updated_nodes = OrderedSet()
        for i, nodes in enumerate(zip(self.inodes, self._inodes_old)):
            if not self.get_manager().same(*nodes):
                updated_nodes.add(i + 1)
                # self.notify_node_updated(i + 1)
        self.get_manager().ref(*filter(None, self.inodes))
        self.get_manager().deref(*filter(None, self._inodes_prev))
        self.get_manager().deref(*filter(None, self._inodes_neg))

        # Only completed nodes should be used for negation in the next stratum.
        self._inodes_old = self.inodes[:]
        self._inodes_prev = [None] * len(self)
        for i, n in enumerate(self.inodes):
            if self._completed[i]:
                self._inodes_prev[i] = n
        self._inodes_neg = [None] * len(self)
        return updated_nodes

    def build_dd(self):
        required_nodes = set([abs(n) for q, n, l in self.labeled() if self.is_probabilistic(n)])
        required_nodes |= set([abs(n) for q, n, v in self.evidence_all() if self.is_probabilistic(n)])

        if self.timeout:
            # signal.signal(signal.SIGALRM, timeout_handler)
            signal.alarm(self.timeout)
            signal.signal(signal.SIGALRM, timeout_handler)
            logging.getLogger('problog').info('Set timeout:', self.timeout)
        try:
            self.init_build()
            updated_nodes = OrderedSet(self._facts)
            while updated_nodes:
                # TODO only check nodes that are actually used in negation
                updated_nodes = self.build_stratum(updated_nodes)
            self._propagate_complete(False)
        except SystemError as err:
            self._propagate_complete(True)
            logging.getLogger('problog').warning(err)
        except KeyboardInterrupt as err:
            self._propagate_complete(True)
            logging.getLogger('problog').warning(err)

        signal.alarm(0)
        self.build_constraint_dd()

    def current(self):
        destination = LogicFormula(auto_compact=False)
        source = self
        # TODO maintain a translation table
        for i, n, t in source:
            inode = self.get_inode(i)
            if inode is not None:
                inode = int(inode)
            if t == 'atom':
                j = destination.add_atom(n.identifier, n.probability, n.group, name=inode)
            elif t == 'conj':
                children = [c for c in n.children if self.get_inode(c) is not None]
                j = destination.add_and(children, name=inode)
            elif t == 'disj':
                children = [c for c in n.children if self.get_inode(c) is not None]
                j = destination.add_or(children, name=inode)
            else:
                raise TypeError('Unknown node type')
            assert i == j

        for name, node, label in source.get_names_with_label():
            if label != self.LABEL_NAMED:
                destination.add_name(name, node, label)

        for c in source.constraints():
            if c.is_nontrivial():
                destination.add_constraint(c)
        return destination

    def update_inode(self, index):
        """Recompute the inode at the given index."""
        oldnode = self.get_inode(index)
        node = self.get_node(index)
        assert index > 0
        nodetype = type(node).__name__
        if nodetype == 'conj':
            children = [self.get_inode(c) for c in node.children]
            children_complete = [self.is_complete(c) for c in node.children]
            if None in children:
                newnode = None  # don't compute if some children are still unknown
            else:
                newnode = self.get_manager().conjoin(*children)
            if False not in children_complete:
                self.set_complete(index)
        elif nodetype == 'disj':
            children = [self.get_inode(c) for c in node.children]
            children_complete = [self.is_complete(c) for c in node.children]
            children = list(filter(None, children))  # discard children that are still unknown
            if children:
                newnode = self.get_manager().disjoin(*children)
            else:
                newnode = None
            if False not in children_complete:
                self.set_complete(index)
        else:
            raise TypeError('Unexpected node type.')

        # Add constraints
        if newnode is not None:
            newernode = self.get_manager().conjoin(newnode, self.get_constraint_inode())
            self.get_manager().deref(newnode)
            newnode = newernode

        if self.get_manager().same(oldnode, newnode):
            return False  # no change occurred
        else:
            if oldnode is not None:
                self.get_manager().deref(oldnode)
            self.set_inode(index, newnode)
            return True

    def get_evidence_inode(self):
        if not self.is_probabilistic(self.evidence_node):
            return self.get_manager().true()
        else:
            inode = self.get_inode(self.evidence_node)
            if inode:
                return inode
            else:
                return self.get_manager().true()

    def get_inode(self, index):
        """
        Get the internal node corresponding to the entry at the given index.
        :param index: index of node to retrieve
        :return: SDD node corresponding to the given index
        :rtype: SDDNode
        """
        assert self.is_probabilistic(index)

        node = self.get_node(abs(index))
        if type(node).__name__ == 'atom':
            av = self.atom2var.get(abs(index))
            if av is None:
                av = self.get_manager().add_variable()
                self.atom2var[abs(index)] = av
            result = self.get_manager().literal(av)
            if index < 0:
                return self.get_manager().negate(result)
            else:
                return result
        elif index < 0:
            # We are requesting a negated node => use previous stratum's result
            result = self._inodes_neg[-index - 1]
            if result is None and self._inodes_prev[-index - 1] is not None:
                result = self.get_manager().negate(self._inodes_prev[-index - 1])
                self._inodes_neg[-index - 1] = result
            return result
        else:
            return self.inodes[index - 1]

    def add_constraint(self, c):
        LogicFormula.add_constraint(self, c)


class _ForwardSDD(SDD, ForwardInference):

    transform_preference = 1000

    def __init__(self, sdd_auto_gc=True, **kwdargs):
        SDD.__init__(self, sdd_auto_gc=sdd_auto_gc, **kwdargs)
        ForwardInference.__init__(self, **kwdargs)

    @classmethod
    def is_available(cls):
        return SDD.is_available()


class _ForwardBDD(BDD, ForwardInference):

    transform_preference = 1000

    def __init__(self, **kwdargs):
        BDD.__init__(self, **kwdargs)
        ForwardInference.__init__(self, **kwdargs)

    @classmethod
    def is_available(cls):
        return BDD.is_available()


@transform(LogicFormula, _ForwardSDD)
def build_sdd(source, destination, **kwdargs):
    result = build_dd(source, destination, **kwdargs)
    return result


@transform(LogicFormula, _ForwardBDD)
def build_sdd(source, destination, **kwdargs):
    result = build_dd(source, destination, **kwdargs)
    return result


class ForwardSDD(LogicFormula, EvaluatableDSP):

    transform_preference = 30

    def __init__(self, **kwargs):
        LogicFormula.__init__(self, **kwargs)
        EvaluatableDSP.__init__(self)
        self.kwargs = kwargs

    @classmethod
    def is_available(cls):
        return SDD.is_available()

    def _create_evaluator(self, semiring, weights, **kwargs):
        return ForwardEvaluator(self, semiring, _ForwardSDD(**self.kwargs), weights, **kwargs)


class ForwardBDD(LogicFormula, EvaluatableDSP):

    transform_preference = 40

    def __init__(self, **kwargs):
        LogicFormula.__init__(self, **kwargs)
        EvaluatableDSP.__init__(self)
        self.kwargs = kwargs

    @classmethod
    def is_available(cls):
        return BDD.is_available()

    def _create_evaluator(self, semiring, weights, **kwargs):
        return ForwardEvaluator(self, semiring, _ForwardBDD(**self.kwargs), weights, **kwargs)


# Inform the system that we can create a ForwardFormula in the same way as a LogicFormula.
transform_create_as(ForwardSDD, LogicFormula)
transform_create_as(ForwardBDD, LogicFormula)


class ForwardEvaluator(Evaluator):
    """An evaluator using anytime forward compilation."""

    def __init__(self, formula, semiring, fdd, weights=None, verbose=None, **kwargs):
        Evaluator.__init__(self, formula, semiring, weights, **kwargs)

        self.fsdd = fdd
        self._z = None
        self._verbose = verbose

        self._results = {}
        self._complete = set()

        self._start_time = None

    def node_updated(self, source, node, complete):

        name = [n for n, i, l in self.formula.labeled()
                if source.is_probabilistic(i) and abs(i) == node]
        if node == abs(source.evidence_node):
            name = ('evidence',)
        if name:
            name = name[0]
            weights = {}
            for atom, weight in self.weights.items():
                av = source.atom2var.get(atom)
                if av is not None:
                    weights[av] = weight
            inode = source.get_inode(node)
            if inode is not None:
                enode = source.get_manager().conjoin(source.get_evidence_inode(),
                                                     source.get_constraint_inode())
                qnode = source.get_manager().conjoin(inode, enode)
                tvalue = source.get_manager().wmc(enode, weights, self.semiring)
                value = source.get_manager().wmc(qnode, weights, self.semiring)
                result = self.semiring.normalize(value, tvalue)
                self._results[node] = result

                debug_msg = 'update query %s: %s after %ss' % \
                            (name, self.semiring.result(result, self.formula),
                             '%.4f' % (time.time() - self._start_time))
                logging.getLogger('problog').debug(debug_msg)

            if complete:
                self._complete.add(node)

    def node_completed(self, source, node):
        qs = set(abs(qi) for qn, qi, ql in source.labeled() if source.is_probabilistic(qi))
        if node in qs:
            self._complete.add(node)

    def initialize(self):
        self.weights = self.formula.extract_weights(self.semiring, self.given_weights)

        # We should do all compilation here.
        self.fsdd.register_update_listener(self)
        self._start_time = time.time()
        build_dd(self.formula, self.fsdd)

        # Update weights with constraints and evidence
        enode = self.fsdd.get_manager().conjoin(self.fsdd.get_evidence_inode(),
                                                self.fsdd.get_constraint_inode())

        # Make sure all atoms exist in atom2var.
        for name, node, label in self.fsdd.labeled():
            if self.fsdd.is_probabilistic(node):
                self.fsdd.get_inode(node)

        weights = {}
        for atom, weight in self.weights.items():
            av = self.fsdd.atom2var.get(atom)
            if av is not None:
                weights[av] = weight

        for name, node, label in self.fsdd.labeled():
            if self.fsdd.is_probabilistic(node):
                inode = self.fsdd.get_inode(node)
                qnode = self.fsdd.get_manager().conjoin(inode, enode)
                tvalue = self.fsdd.get_manager().wmc(enode, weights, self.semiring)
                value = self.fsdd.get_manager().wmc(qnode, weights, self.semiring)
                result = self.semiring.normalize(value, tvalue)
            elif self.fsdd.is_true(node):
                result = self.semiring.one()
            else:
                result = self.semiring.zero()
            self._results[node] = result


    def propagate(self):
        self.initialize()

    def evaluate(self, index):
        """Compute the value of the given node."""
        # We should get results from cache here.

        ub = 1.0
        if index is None:
            return 0.0
        elif index == 0:
            return 1.0
        else:
            n = self.formula.get_node(abs(index))
            nt = type(n).__name__
            if nt == 'atom':
                wp = self._results[abs(index)]
                # wp, wn = self.weights.get(abs(index))
                if index < 0:
                    wn = self.semiring.negate(wp)
                    return self.semiring.result(wn, self.formula)
                else:
                    return self.semiring.result(wp, self.formula)
            else:
                # TODO report correct bounds in case of evidence
                if index < 0:
                    if -index in self._results:
                        if -index in self._complete:
                            return self.semiring.result(self.semiring.negate(self._results[-index]),
                                                        self.formula)
                        else:
                            return 0.0, self.semiring.result(
                                self.semiring.negate(self._results[-index]), self.formula)
                    else:
                        return 0.0, 1.0
                else:
                    if index in self._results:
                        if index in self._complete:
                            return self.semiring.result(self._results[index], self.formula)
                        else:
                            return self.semiring.result(self._results[index], self.formula), 1.0
                    else:
                        return 0.0, 1.0

    def evaluate_evidence(self):
        raise NotImplementedError('Evaluator.evaluate_evidence is an abstract method.')

    # def add_evidence(self, node):
    #     """Add evidence"""
    #     warnings.warn('Evidence is not supported by this evaluation method and will be ignored.')

    def has_evidence(self):
        return self.__evidence != []

    def clear_evidence(self):
        self.__evidence = []

    def evidence(self):
        return iter(self.__evidence)



#! /usr/bin/env python
"""
Part of the ProbLog distribution.

Copyright 2015 KU Leuven, DTAI Research Group

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
from __future__ import print_function

import os, sys, subprocess, traceback, json

sys.setrecursionlimit(10000)

sys.path.insert(0, os.path.abspath( os.path.join(os.path.dirname(__file__), '../')))

from problog.program import PrologFile
from problog.errors import process_error

from learning import lfi

# from problog.sdd_formula import SDD

def print_result( d, output, precision=8 ) :
    success, d = d
    if success :
        score, weights, names, iterations = d

        results = { 'score' : score, 'iterations' : iterations, 'weights': [[n[0],w,n[1],n[2]] for n,w in zip(names,weights)] }
        results['SUCCESS'] = True
        print (200, 'application/json', json.dumps(results), file=output)
    else :
        d['SUCCESS'] = False
        #print (400, 'application/json', json.dumps(d), file=output)
        print (200, 'application/json', json.dumps(d), file=output)
    return 0


def main(filename, examplefile) :
    try :
        examples = list(lfi.read_examples( examplefile ))
        program = PrologFile(filename)
        score, weights, names, iterations = lfi.run_lfi( program, examples)

        new_names = []
        for n in names:
            new_names.append((str(n.with_probability()),) + program.lineno(n.location)[1:])

        return True, (score, weights, new_names, iterations)
    except Exception as err :
        return False, {'err':process_error(err)}


if __name__ == '__main__' :
    import argparse

    parser = argparse.ArgumentParser()
    parser.add_argument('filename', metavar='MODEL')
    parser.add_argument('examples', metavar='EXAMPLES')
    parser.add_argument('output', metavar='OUTPUT')

    args = parser.parse_args()

    result = main(args.filename, args.examples)
    with open(args.output, 'w') as output :
        retcode = print_result( result , output )

    # Always exit with code 0



"""

..
    Part of the ProbLog distribution.

    Copyright 2015 KU Leuven, DTAI Research Group

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

from __future__ import print_function

from ..program import PrologFile
from ..engine import DefaultEngine
from ..util import init_logger, format_dictionary
from ..kbest import KBestFormula
from ..errors import process_error

import argparse
import sys
import json
import traceback


def main(argv):

    parser = argparse.ArgumentParser()
    parser.add_argument('filename')
    parser.add_argument('-v', '--verbose', action='count')
    parser.add_argument('--web', action='store_true')
    parser.add_argument('-o', '--output', type=str, default=None)
    args = parser.parse_args(argv)

    init_logger(args.verbose)
    if args.output:
        out = open(args.output, 'w')
    else:
        out = sys.stdout

    try:
        pl = PrologFile(args.filename)
        db = DefaultEngine().prepare(pl)

        program = list(map(lambda s: '%s.' % s, db.iter_raw()))
        cnf = KBestFormula.create_from(db, label_all=True)

        explanation = []
        results = cnf.evaluate(explain=explanation)

        if args.web:
            result = {}
            result['SUCCESS'] = True
            result['program'] = program
            result['proofs'] = explanation
            result['probabilities'] = [(str(k), round(v, 8)) for k, v in results.items()]
            print (json.dumps(result), file=out)
        else:
            print ('Transformed program', file=out)
            print ('-------------------', file=out)
            print ('\n'.join(program), file=out)
            print (file=out)

            print ('Proofs', file=out)
            print ('------', file=out)
            print ('\n'.join(explanation), file=out)

            print (file=out)
            print ('Probabilities', file=out)
            print ('-------------', file=out)
            print (format_dictionary(results), file=out)
    except Exception as err:
        trace = traceback.format_exc()
        err.trace = trace
        result = {}
        result['SUCCESS'] = False
        result['err'] = vars(err)
        result['err']['message'] = process_error(err)
        print (json.dumps(result), file=out)

    if args.output:
        out.close()


if __name__ == '__main__':
    main(sys.argv[1:])


def argparser():
    import argparse
    parser = argparse.ArgumentParser()
    parser.add_argument('inputfile')
    return parser


if __name__ == '__main__':
    main(**vars(argparser().parse_args()))


"""
problog.nnf_formula - d-DNNF
----------------------------

Provides access to d-DNNF formulae.

..
    Part of the ProbLog distribution.

    Copyright 2015 KU Leuven, DTAI Research Group

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""
from __future__ import print_function

import tempfile
import os
import subprocess
from collections import defaultdict

from . import system_info
from .evaluator import Evaluator, EvaluatableDSP
from .errors import InconsistentEvidenceError
from .formula import LogicDAG
from .cnf_formula import CNF
from .core import transform
from .errors import CompilationError
from .util import Timer, subprocess_check_call


class DSharpError(CompilationError):
    """DSharp has crashed."""

    def __init__(self):
        msg = 'DSharp has encountered an error'
        if system_info['os'] == 'darwin':
            msg += '. This is a known issue. See KNOWN_ISSUES for details on how to resolve this problem'
        CompilationError.__init__(self, msg)


class NNF(LogicDAG, EvaluatableDSP):
    """A d-DNNF formula."""

    transform_preference = 20

    # noinspection PyUnusedLocal,PyUnusedLocal,PyUnusedLocal
    def __init__(self, **kwdargs):
        LogicDAG.__init__(self, auto_compact=False)

    def _create_evaluator(self, semiring, weights, **kwargs):
        return SimpleNNFEvaluator(self, semiring, weights)


class SimpleNNFEvaluator(Evaluator):
    """Evaluator for d-DNNFs."""

    def __init__(self, formula, semiring, weights=None, **kwargs):
        Evaluator.__init__(self, formula, semiring, weights, **kwargs)

    def _initialize(self, with_evidence=True):
        self.weights.clear()

        model_weights = self.formula.extract_weights(self.semiring, self.given_weights)
        for n, p in model_weights.items():
            self.weights[n] = p[0]
            self.weights[-n] = p[1]

        if with_evidence:
            for ev in self.evidence():
                self.set_evidence(abs(ev), ev > 0)

        if self.semiring.result(self._get_z(), self.formula) == 0.0:
            raise InconsistentEvidenceError(context=" during evidence evaluation")

    def propagate(self):
        self._initialize()

    def _get_z(self):
        result = self._get_weight(len(self.formula))
        return result

    def evaluate_evidence(self, recompute=False):
        return self.semiring.result(self._evaluate_evidence(recompute=recompute), self.formula)

    # noinspection PyUnusedLocal
    def _evaluate_evidence(self, recompute=False):
        self._initialize(False)
        for ev in self.evidence():
            self._set_value(abs(ev), ev > 0)

        result = self._get_weight(len(self.formula))
        return result

    def evaluate_fact(self, node):
        return self.evaluate(node)

    def evaluate(self, node):
        if node == 0:
            result = self.semiring.one()
        elif node is None:
            result = self.semiring.zero()
        else:
            p = self._get_weight(abs(node))
            n = self._get_weight(-abs(node))
            self._set_value(abs(node), (node > 0))
            result = self._get_weight(len(self.formula))
            self._reset_value(abs(node), p, n)
            if self.has_evidence():
                result = self.semiring.normalize(result, self._get_z())
        return self.semiring.result(result, self.formula)

    def _reset_value(self, index, pos, neg):
        self.set_weight(index, pos, neg)

    def _get_weight(self, index):
        if index == 0:
            return self.semiring.one()
        elif index is None:
            return self.semiring.zero()
        else:
            w = self.weights.get(index)
            if w is None:
                w = self._calculate_weight(index)
                return w
            else:
                return w

    def set_weight(self, index, pos, neg):
        self.weights[index] = pos
        self.weights[-index] = neg

    def set_evidence(self, index, value):
        pos = self.semiring.one()
        neg = self.semiring.zero()

        current_weight = self.weights.get(index)
        neg_weight = self.weights.get(-index)

        if value:
            if current_weight is not None and self.semiring.is_zero(current_weight):
                raise InconsistentEvidenceError(self._deref_node(index))
            self.set_weight(index, pos, neg)
        else:
            if current_weight is not None and self.semiring.is_one(current_weight) \
                    and not self.semiring.is_one(neg_weight):
                raise InconsistentEvidenceError(self._deref_node(index))
            self.set_weight(index, neg, pos)

    def _deref_node(self, index):
        return self.formula.get_node(index).name

    def _set_value(self, index, value):
        """Set value for given node.

        :param index: index of node
        :param value: value
        """
        if value:
            pos = self._get_weight(index)
            self.set_weight(index, pos, self.semiring.zero())
        else:
            neg = self._get_weight(-index)
            self.set_weight(index, self.semiring.zero(), neg)

    def _calculate_weight(self, key):
        assert (key != 0)
        assert (key is not None)
        # assert(key > 0)

        node = self.formula.get_node(abs(key))
        ntype = type(node).__name__

        if ntype == 'atom':
            return self.semiring.one()
        else:
            assert key > 0
            childprobs = [self._get_weight(c) for c in node.children]
            if ntype == 'conj':
                p = self.semiring.one()
                for c in childprobs:
                    p = self.semiring.times(p, c)
                return p
            elif ntype == 'disj':
                p = self.semiring.zero()
                for c in childprobs:
                    p = self.semiring.plus(p, c)
                return p
            else:
                raise TypeError("Unexpected node type: '%s'." % ntype)


class Compiler(object):
    """Interface to CNF to d-DNNF compiler tool."""
    __compilers = {}

    @classmethod
    def get_default(cls):
        """Get default compiler for this system."""
        if system_info.get('c2d', False):
            return _compile_with_c2d
        else:
            return _compile_with_dsharp

    @classmethod
    def get(cls, name):
        """Get compiler by name (or default if name not found).

        :param name: name of the compiler
        :returns: function used to call compiler
        """
        result = cls.__compilers.get(name)
        if result is None:
            result = cls.get_default()
        return result

    @classmethod
    def add(cls, name, func):
        """Add a compiler.

        :param name: name of the compiler
        :param func: function used to call the compiler
        """
        cls.__compilers[name] = func


if system_info.get('c2d', False):
    # noinspection PyUnusedLocal
    @transform(CNF, NNF)
    def _compile_with_c2d(cnf, nnf=None, smooth=True, **kwdargs):
        fd, cnf_file = tempfile.mkstemp('.cnf')
        os.close(fd)
        nnf_file = cnf_file + '.nnf'
        if smooth:
            smoothl = ['-smooth_all']
        else:
            smoothl = []

        cmd = ['cnf2dDNNF', '-dt_method', '0'] + smoothl + ['-reduce', '-in', cnf_file]

        try:
            os.remove(cnf_file)
        except OSError:
            pass
        try:
            os.remove(nnf_file)
        except OSError:
            pass

        return _compile(cnf, cmd, cnf_file, nnf_file)


    Compiler.add('c2d', _compile_with_c2d)


# noinspection PyUnusedLocal
@transform(CNF, NNF)
def _compile_with_dsharp(cnf, nnf=None, smooth=True, **kwdargs):
    result = None
    with Timer('DSharp compilation'):
        fd1, cnf_file = tempfile.mkstemp('.cnf')
        fd2, nnf_file = tempfile.mkstemp('.nnf')
        os.close(fd1)
        os.close(fd2)
        if smooth:
            smoothl = ['-smoothNNF']
        else:
            smoothl = []
        cmd = ['dsharp', '-Fnnf', nnf_file] + smoothl + ['-disableAllLits', cnf_file]  #

        try:
            result = _compile(cnf, cmd, cnf_file, nnf_file)
        except subprocess.CalledProcessError:
            raise DSharpError()

        try:
            os.remove(cnf_file)
        except OSError:
            pass
        try:
            os.remove(nnf_file)
        except OSError:
            pass

    return result


Compiler.add('dsharp', _compile_with_dsharp)


def _compile(cnf, cmd, cnf_file, nnf_file):
    names = cnf.get_names_with_label()

    if cnf.is_trivial():
        nnf = NNF()
        weights = cnf.get_weights()
        for i in range(1, cnf.atomcount + 1):
            nnf.add_atom(i, weights.get(i))
        or_nodes = []
        for i in range(1, cnf.atomcount + 1):
            or_nodes.append(nnf.add_or((i, -i)))
        if or_nodes:
            nnf.add_and(or_nodes)

        for name, node, label in names:
            nnf.add_name(name, node, label)
        for c in cnf.constraints():
            nnf.add_constraint(c.copy())

        return nnf
    else:
        with open(cnf_file, 'w') as f:
            f.write(cnf.to_dimacs())

        attempts_left = 1
        success = False
        while attempts_left and not success:
            try:
                with open(os.devnull, 'w') as OUT_NULL:
                    subprocess_check_call(cmd, stdout=OUT_NULL)
                success = True
            except subprocess.CalledProcessError as err:
                attempts_left -= 1
                if attempts_left == 0:
                    raise err
        return _load_nnf(nnf_file, cnf)


def _load_nnf(filename, cnf):
    nnf = NNF()

    weights = cnf.get_weights()

    names_inv = defaultdict(list)
    for name, node, label in cnf.get_names_with_label():
        names_inv[node].append((name, label))

    with open(filename) as f:
        line2node = {}
        rename = {}
        lnum = 0
        for line in f:
            line = line.strip().split()
            if line[0] == 'nnf':
                pass
            elif line[0] == 'L':
                name = int(line[1])
                prob = weights.get(abs(name), True)
                node = nnf.add_atom(abs(name), prob)
                rename[abs(name)] = node
                if name < 0:
                    node = -node
                line2node[lnum] = node
                if name in names_inv:
                    for actual_name, label in names_inv[name]:
                        nnf.add_name(actual_name, node, label)
                    del names_inv[name]
                lnum += 1
            elif line[0] == 'A':
                children = map(lambda x: line2node[int(x)], line[2:])
                line2node[lnum] = nnf.add_and(children)
                lnum += 1
            elif line[0] == 'O':
                children = map(lambda x: line2node[int(x)], line[3:])
                line2node[lnum] = nnf.add_or(children)
                lnum += 1
            else:
                print('Unknown line type')
        for name in names_inv:
            for actual_name, label in names_inv[name]:
                if name == 0:
                    nnf.add_name(actual_name, 0, label)
                else:
                    nnf.add_name(actual_name, None, label)
    for c in cnf.constraints():
        nnf.add_constraint(c.copy(rename))

    return nnf



#! /usr/bin/env python

"""
Learning from interpretations
-----------------------------

Parameter learning for ProbLog.

Given a probabilistic program with parameterized weights and a set of partial implementations,
learns appropriate values of the parameters.

Algorithm
+++++++++

The algorithm operates as follows:

    0. Set initial values for the weights to learn.
    1. Set the evidence present in the example.
    2. Query the model for the weights of the atoms to be learned.
    3. Update the weights to learn by taking the mean value over all examples and queries.
    4. Repeat steps 1 to 4 until convergence (or a maximum number of iterations).
    
The score of the model for a given example is obtained by calculating the probability of the
evidence in the example.

Implementation
++++++++++++++

The algorithm is implemented on top of the ProbLog toolbox.

It uses the following extensions of ProbLog's classes:

    * a LogicProgram implementation that rewrites the model and extracts the weights to learn
    (see :py:func:`learning.lfi.LFIProblem.__iter__`)
    * a custom semiring that looks up the current value of a weight to learn
    (see :py:func:`learning.lfi.LFIProblem.value`)


.. autoclass:: learning.lfi.LFIProblem
    :members: __iter__, value

"""

from __future__ import print_function

import sys
import random
import math
import logging

from collections import defaultdict

from problog.engine import DefaultEngine, ground
from problog.evaluator import SemiringProbability
from problog.logic import Term, Constant, Clause, AnnotatedDisjunction, Or, Var,\
    InstantiationError, ArithmeticError
from problog.program import PrologString, PrologFile, LogicProgram
from problog.core import ProbLogError
from problog.errors import process_error, InconsistentEvidenceError
from problog.sdd_formula import SDD

from problog import get_evaluatable, get_evaluatables
import traceback


def str2bool(s):
    if str(s) == 'true':
        return True
    elif str(s) == 'false':
        return False
    else:
        return None


class LFIProblem(SemiringProbability, LogicProgram):

    def __init__(self, source, examples, max_iter=10000, min_improv=1e-10, verbose=0, knowledge=SDD,
                 leakprob=None, propagate_evidence=True, normalize=False, **extra):
        """
        :param source: filename of file containing input model
        :type source: str
        :param examples: list of observed terms / value
        :type examples: list[tuple(Term, bool)]
        :param max_iter: maximum number of iterations to run
        :type max_iter: int
        :param min_improv: minimum improvement in log-likelihood for convergence detection
        :type min_improv: float
        :param verbose: verbosity level
        :type verbose: int
        :param knowledge: class to use for knowledge compilation
        :type knowledge: class
        :param leakprob: Add all true evidence atoms with the given probability
                         to avoid 'inconsistent evidence' errors. This also
                         allows to learn a program without constants and
                         retrieve the constants from the evidence file.
                         (default: None)
        :type leakprob: float or None
        :param extra: catch all for additional parameters (not used)
        """
        SemiringProbability.__init__(self)
        LogicProgram.__init__(self)
        self.source = source

        # The names of the atom for which we want to learn weights.
        self.names = []

        # The weights to learn.
        # The initial weights are of type 'float'.
        # When necessary they are replaced by a dictionary [t(arg1, arg2, ...) -> float]
        #  for weights of form t(SV, arg1, arg2, ...).
        self._weights = []

        self.examples = examples
        self.leakprob = leakprob
        self.leakprobatoms = None
        self.propagate_evidence = propagate_evidence
        self._compiled_examples = None
        
        self.max_iter = max_iter
        self.min_improv = min_improv
        self.verbose = verbose
        self.iteration = 0

        self.knowledge = knowledge

        self.output_mode = False
        self.extra = extra

        self._enable_normalize = normalize
        self._adatoms = []
    
    def value(self, a):
        """Overrides from SemiringProbability.
        Replaces a weight of the form ``lfi(i, t(...))`` by its current estimated value.
        Other weights are passed through unchanged.

        :param a: term representing the weight
        :type a: Term
        :return: current weight
        :rtype: float
        """
        if isinstance(a, Term) and a.functor == 'lfi':
            # index = int(a.args[0])
            return self._get_weight(*a.args)
        else:
            return float(a)
         
    @property 
    def count(self):
        """Number of parameters to learn."""
        return len(self.names)
    
    def prepare(self):
        """Prepare for learning."""
        self._compile_examples()

    def _get_weight(self, index, args, strict=True):
        index = int(index)
        weight = self._weights[index]
        if isinstance(weight, dict):
            if strict:
                return weight[args]
            else:
                return weight.get(args, 0.0)
        else:
            return weight

    def get_weights(self, index):
        """Get a list of key, weight pairs for the given input fact.

        :param index: identifier of the fact
        :return: list of key, weight pairs where key refers to the additional variables
        on which the weight is based
        :rtype: list[tuple[Term, float]]
        """
        weight = self._weights[index]
        if isinstance(weight, dict):
            return list(weight.items())
        else:
            return [(Term('t'), weight)]

    def _set_weight(self, index, args, weight):
        index = int(index)
        if not args:
            assert not isinstance(self._weights[index], dict)
            self._weights[index] = weight
        elif isinstance(self._weights[index], dict):
            self._weights[index][args] = weight
        else:
            self._weights[index] = {args: weight}

    def _add_weight(self, weight):
        self._weights.append(weight)

    def _process_examples(self):
        """Process examples by grouping together examples with similar structure.
    
        :return: example groups based on evidence atoms
        :rtype: dict of atoms : values for examples
        """
    
        # value can be True / False / None
        # ( atom ), ( ( value, ... ), ... ) 

        # Simple implementation: don't add neutral evidence.

        if self.propagate_evidence:
            result = []
            for example in self.examples:
                atoms, values = zip(*example)
                result.append((atoms, [values]))
            return result
        else:
            result = defaultdict(list)
            for example in self.examples:
                atoms, values = zip(*example)
                result[atoms].append(values)
            return result.items()
    
    def _compile_examples(self):
        """Compile examples.
    
        :param examples: Output of ::func::`process_examples`.
        """
        logger = logging.getLogger('problog_lfi')

        baseprogram = DefaultEngine(**self.extra).prepare(self)
        examples = self._process_examples()
        result = []
        for atoms, example_group in examples:
            ground_program = None   # Let the grounder decide
            for n, example in enumerate(example_group):
                if self.verbose:
                    logger.debug('Compiling example %s ...' % n)

                ground_program = ground(baseprogram, ground_program,
                                        evidence=list(zip(atoms, example)),
                                        propagate_evidence=self.propagate_evidence)
                for i, node, t in ground_program:
                    if t == 'atom' and \
                            isinstance(node.probability, Term) and \
                            node.probability.functor == 'lfi':
                        factargs = ()
                        if type(node.identifier) == tuple:
                            factargs = node.identifier[1]
                        fact = Term('lfi_fact', node.probability.args[0], node.probability.args[1], *factargs)
                        ground_program.add_query(fact, i)
                compiled_program = self.knowledge.create_from(ground_program)
                result.append((atoms, example, compiled_program, n))
        self._compiled_examples = result

    def _process_atom(self, atom, body):
        """Returns tuple ( prob_atom, [ additional clauses ] )"""
        if isinstance(atom, Or):
            # Annotated disjunction
            atoms = atom.to_list()
        else:
            atoms = [atom]

        atoms_out = []
        extra_clauses = []

        has_lfi_fact = False
        available_probability = 1.0

        num_random_weights = 0
        for atom in atoms:
            if atom.probability and atom.probability.functor == 't':
                try:
                    start_value = float(atom.probability.args[0])
                    available_probability -= float(start_value)
                except InstantiationError:
                    # Can't be converted to float => take random
                    num_random_weights += 1
                except ArithmeticError:
                    num_random_weights += 1

            elif atom.probability:
                available_probability -= float(atom.probability)

        random_weights = [random.random() for i in range(0, num_random_weights + 1)]
        norm_factor = available_probability / sum(random_weights)
        random_weights = [r * norm_factor for r in random_weights]

        self._adatoms.append((available_probability, []))

        for atom in atoms:
            if atom.probability and atom.probability.functor == 't':
                # t(_)::p(X) :- body.
                #
                # Translate to
                #   lfi(1)::lfi_fact_1(X).
                #   p(X) :- body, lfi_fact1(X).
                # For annotated disjunction: t(_)::p1(X); t(_)::p2(X) :- body.
                #   lfi1::lfi_fact1(X); lfi2::lfi_fact2(X); ... .
                #   p1(X) :- body, lfi_fact1(X).
                #   p2(X) :- body, lfi_fact2(X).
                #  ....
                has_lfi_fact = True

                # Learnable probability
                try:
                    start_value = float(atom.probability.args[0])
                except InstantiationError:
                    start_value = None
                except ArithmeticError:
                    start_value = None

                # Replace anonymous variables with non-anonymous variables.
                class ReplaceAnon(object):

                    def __init__(self):
                        self.cnt = 0

                    def __getitem__(self, key):
                        if key == '_':
                            self.cnt += 1
                            return Var('anon_%s' % self.cnt)
                        else:
                            return Var(key)

                atom1 = atom.apply(ReplaceAnon())
                prob_args = atom.probability.args[1:]

                # 1) Introduce a new fact
                lfi_fact = Term('lfi_fact', Constant(self.count), Term('t', *prob_args), *atom1.args)
                lfi_prob = Term('lfi', Constant(self.count), Term('t', *prob_args))

                # 2) Replacement atom
                replacement = lfi_fact.with_probability(lfi_prob)
                if body is None:
                    new_body = lfi_fact
                else:
                    new_body = body & lfi_fact

                # 3) Create redirection clause
                extra_clauses += [Clause(atom1.with_probability(), new_body)]

                self._adatoms[-1][1].append(len(self._weights))
                # 4) Set initial weight
                if start_value is None:
                    self._add_weight(random_weights.pop(-1))
                else:
                    self._add_weight(start_value)

                # 5) Add name
                self.names.append(atom)
                atoms_out.append(replacement)
            else:
                atoms_out.append(atom)

        if len(self._adatoms[-1][1]) < 2:
            self._adatoms.pop(-1)

        if has_lfi_fact:
            if len(atoms) == 1:     # Simple clause
                return [atoms_out[0]] + extra_clauses
            else:
                return [AnnotatedDisjunction(atoms_out, Term('true'))] + extra_clauses
        else:
            if len(atoms) == 1:
                if body is None:
                    return [atoms_out[0]]
                else:
                    return [Clause(atoms_out[0], body)]
            else:
                if body is None:
                    body = Term('true')
                return [AnnotatedDisjunction(atoms_out, body)]

    def _process_atom_output(self, atom, body):
        """Returns tuple ( prob_atom, [ additional clauses ] )"""

        if isinstance(atom, Or):
            atoms = atom.to_list()
        else:
            atoms = [atom]

        transforms = defaultdict(list)

        clauses = []
        atoms_fixed = []
        t_args = None
        fixed_only = True
        for atom in atoms:
            if atom.probability and atom.probability.functor == 't':
                assert (atom in self.names)
                # assert (t_args is None or atom.probability.args == t_args)
                # t_args = atom.probability.args

                index = self.output_names.index(atom)
                weights = self.get_weights(index)

                for w_args, w_val in weights:
                    translate = tuple(zip(atom.probability.args[1:], w_args.args))
                    transforms[translate].append(atom.with_probability(Constant(w_val)))
                self.output_names[index] = None
                fixed_only = False
            else:
                atoms_fixed.append(atom)

        if not fixed_only:
            clauses = []
            for tr, atoms in transforms.items():
                tr = DefaultDict({k: v for k, v in tr})
                atoms_out = [at.apply(tr) for at in atoms] + atoms_fixed
                if len(atoms_out) == 1:
                    if body is None:
                        clauses.append(atoms_out[0])
                    else:
                        clauses.append(Clause(atoms_out[0], body.apply(tr)))
                else:
                    if body is None:
                        clauses.append(AnnotatedDisjunction(atoms_out, None))
                    else:
                        clauses.append(AnnotatedDisjunction(atoms_out, body.apply(tr)))
            return clauses
        else:
            atoms_out = atoms_fixed
            if len(atoms_out) == 1:
                if body is None:
                    return [atoms_out[0]]
                else:
                    return [Clause(atoms_out[0], body)]
            else:
                return [AnnotatedDisjunction(atoms_out, body)]

    # Overwrite from LogicProgram    
    def __iter__(self):
        """
        Iterate over the clauses of the source model.
        This object can be used as a LogicProgram to be passed to the grounding Engine.
        
        Extracts and processes all ``t(...)`` weights.
        This
        
            * replaces each probabilistic atom ``t(...)::p(X)`` by a unique atom \
            ``lfi(i) :: lfi_fact_i(X)``;
            * adds a new clause ``p(X) :- lfi_fact_i(X)``;
            * adds a new query ``query( lfi_fact_i(X) )``;
            * initializes the weight of ``lfi(i)`` based on the ``t(...)`` specification;
        
        This also removes all existing queries from the model.
        
        Example:
        
        .. code-block:: prolog
        
            t(_) :: p(X) :- b(X).
            t(_) :: p(X) :- c(X).

        is transformed into
        
        .. code-block:: prolog
        
            lfi(0) :: lfi_fact_0(X) :- b(X).
            p(X) :- lfi_fact_0(X).
            lfi(1) :: lfi_fact_1(X) :- c(X).
            p(X) :- lfi_fact_1(X).
            query(lfi_fact_0(X)).
            query(lfi_fact_1(X)).

        If ``self.leakprobs`` is a value, then during learning all true
        examples are added to the program with the given leak probability.

        """

        if self.output_mode:
            process_atom = self._process_atom_output
            self.output_names = self.names[:]
        else:
            process_atom = self._process_atom

        for clause in self.source:
            if isinstance(clause, Clause):
                if clause.head.functor == 'query' and clause.head.arity == 1:
                    continue
                extra_clauses = process_atom(clause.head, clause.body)
                for extra in extra_clauses:
                    yield extra
            elif isinstance(clause, AnnotatedDisjunction):
                extra_clauses = process_atom(Or.from_list(clause.heads), clause.body)
                for extra in extra_clauses:
                    yield extra
            else:
                if clause.functor == 'query' and clause.arity == 1:
                    continue
                # Fact
                extra_clauses = process_atom(clause, None)
                for extra in extra_clauses:
                    yield extra

        if self.leakprob is not None:
            leakprob_atoms = self._get_leakprobatoms()
            for example_atom in leakprob_atoms:
                yield example_atom.with_probability(Constant(self.leakprob))

    def _get_leakprobatoms(self):
        if self.leakprobatoms is not None:
            return self.leakprobatoms
        self.leakprobatoms = set()
        for examples in self.examples:
            for example, obs in examples:
                if obs:
                    self.leakprobatoms.add(example)
        return self.leakprobatoms

    def _evaluate_examples(self):
        """Evaluate the model with its current estimates for all examples."""
        
        results = []
        i = 0
        logging.getLogger('problog_lfi').debug('Evaluating examples ...')
        for at, val, comp, n in self._compiled_examples:
            evidence = {}
            for a, v in zip(at, val):
                if a in evidence:
                    if evidence[a] != v:
                        context = ' (found evidence({},{}) and evidence({},{}) in example {})'.format(a, evidence[a], a, v, n+1)
                        raise InconsistentEvidenceError(source=a, context=context)
                else:
                    evidence[a] = v
            try:
                evaluator = comp.get_evaluator(semiring=self, evidence=evidence)
            except InconsistentEvidenceError as err:
                if err.context == '':
                    context = ' (example {})'.format(n+1)
                else:
                    context = err.context + ' (example {})'.format(n+1)
                raise InconsistentEvidenceError(err.source, context)
            p_queries = {}
            # Probability of query given evidence
            for name, node, label in evaluator.formula.labeled():
                w = evaluator.evaluate_fact(node)
                if w < 1e-6:
                    p_queries[name] = 0.0
                else:
                    p_queries[name] = w
            p_evidence = evaluator.evaluate_evidence()
            i += 1
            results.append((p_evidence, p_queries))
        return results
    
    def _update(self, results):
        """Update the current estimates based on the latest evaluation results."""
        
        fact_marg = defaultdict(float)
        fact_count = defaultdict(int)
        score = 0.0
        for pEvidence, result in results:
            for fact, value in result.items():
                index = fact.args[0:2]
                fact_marg[index] += value
                fact_count[index] += 1
            try:
                score += math.log(pEvidence)
            except ValueError:
                raise ProbLogError('Inconsistent evidence.')

        for index in fact_marg:
            if fact_count[index] > 0:
                self._set_weight(index[0], index[1], fact_marg[index] / fact_count[index])

        if self._enable_normalize:
            self._normalize_weights()
        return score

    def _normalize_weights(self):

        for p, idx in self._adatoms:
            keys = set()
            for i in idx:
                for key, val in self.get_weights(i):
                    keys.add(key)
            if len(keys) > 1:
                try:
                    keys.remove(Term('t'))
                except KeyError:
                    pass
            for key in keys:
                w = sum(self._get_weight(i, key, strict=False) for i in idx)
                n = p / w
                for i in idx:
                    self._set_weight(i, key, self._get_weight(i, key, strict=False) * n)
        
    def step(self):
        self.iteration += 1
        results = self._evaluate_examples()
        return self._update(results)

    def get_model(self):
        self.output_mode = True
        lines = []
        for l in self:
            lines.append('%s.' % l)
        lines.append('')
        self.output_mode = False
        return '\n'.join(lines)
        
    def run(self):
        self.prepare()
        logging.getLogger('problog_lfi').info('Weights to learn: %s' % self.names)
        logging.getLogger('problog_lfi').info('Initial weights: %s' % self._weights)
        delta = 1000
        prev_score = -1e10
        while self.iteration < self.max_iter and (delta < 0 or delta > self.min_improv):
            score = self.step()
            logging.getLogger('problog_lfi').info('Weights after iteration %s: %s' % (self.iteration, self._weights))
            logging.getLogger('problog_lfi').info('Score after iteration %s: %s' % (self.iteration, score))
            delta = score - prev_score
            prev_score = score
        return prev_score


def extract_evidence(pl):
    engine = DefaultEngine()
    atoms = engine.query(pl, Term('evidence', None, None))
    atoms1 = engine.query(pl, Term('evidence', None))
    atoms2 = engine.query(pl, Term('observe', None))
    for atom in atoms1 + atoms2:
        atom = atom[0]
        if atom.is_negated():
            atoms.append((-atom, Term('false')))
        else:
            atoms.append((atom, Term('true')))
    return [(at, str2bool(vl)) for at, vl in atoms]


def read_examples(*filenames):
    
    for filename in filenames:
        engine = DefaultEngine()
        
        with open(filename) as f:
            example = ''
            for line in f:
                if line.strip().startswith('---'):
                    pl = PrologString(example)
                    atoms = extract_evidence(pl)
                    if len(atoms) > 0:
                        yield atoms
                    example = ''
                else:
                    example += line
            if example:
                pl = PrologString(example)
                atoms = extract_evidence(pl)
                if len(atoms) > 0:
                    yield atoms


class DefaultDict(object):

    def __init__(self, base):
        self.base = base

    def __getitem__(self, key):
        return self.base.get(key, Var(key))

    
def run_lfi(program, examples, output_model=None, **kwdargs):
    lfi = LFIProblem(program, examples, **kwdargs)
    score = lfi.run()

    if output_model is not None:
        with open(output_model, 'w') as f:
            f.write(lfi.get_model())

    names = []
    weights = []
    for i, name in enumerate(lfi.names):
        weights_i = lfi.get_weights(i)
        for w_args, w_val in weights_i:
            translate = {k: v for k, v in zip(name.probability.args[1:], w_args.args)}
            names.append(name.apply(DefaultDict(translate)))
            weights.append(w_val)

    return score, weights, names, lfi.iteration, lfi


def argparser():
    import argparse
    parser = argparse.ArgumentParser(description="Learning from interpretations with ProbLog")
    parser.add_argument('model')
    parser.add_argument('examples', nargs='+')
    parser.add_argument('-n', dest='max_iter', default=10000, type=int )
    parser.add_argument('-d', dest='min_improv', default=1e-10, type=float )
    parser.add_argument('-O', '--output-model', type=str, default=None,
                        help='write resulting model to given file')
    parser.add_argument('-o', '--output', type=str, default=None,
                        help='write output to file')
    parser.add_argument('-k', '--knowledge', dest='koption', choices=get_evaluatables(),
                        default=None, help='knowledge compilation tool')
    parser.add_argument('-l', '--leak-probabilities', dest='leakprob', type=float,
                        help='Add leak probabilities for evidence atoms.')
    parser.add_argument('--propagate-evidence', action='store_true',
                        dest='propagate_evidence',
                        default=True,
                        help="Enable evidence propagation")
    parser.add_argument('--dont-propagate-evidence', action='store_false',
                        dest='propagate_evidence',
                        default=True,
                        help="Disable evidence propagation")
    parser.add_argument('--normalize', action='store_true', help="Normalize AD-weights.")
    parser.add_argument('-v', '--verbose', action='count', default=0)
    parser.add_argument('--web', action='store_true', help=argparse.SUPPRESS)
    parser.add_argument('-a', '--arg', dest='args', action='append',
                        help='Pass additional arguments to the cmd_args builtin.')

    return parser


def create_logger(name, verbose):
    levels = [logging.WARNING, logging.INFO, logging.DEBUG] + list(range(9, 0, -1))
    verbose = max(0, min(len(levels)-1, verbose))
    logger = logging.getLogger(name)
    ch = logging.StreamHandler(sys.stdout)
    formatter = logging.Formatter('[%(levelname)s] %(message)s')
    ch.setFormatter(formatter)
    logger.addHandler(ch)
    logger.setLevel(levels[verbose])


def main(argv, result_handler=None):
    parser = argparser()
    args = parser.parse_args(argv)

    if result_handler is None:
        if args.web:
            result_handler = print_result_json
        else:
            result_handler = print_result

    knowledge = get_evaluatable(args.koption)

    if args.output is None:
        outf = sys.stdout
    else:
        outf = open(args.output, 'w')

    create_logger('problog_lfi', args.verbose)
    create_logger('problog', args.verbose - 1)

    program = PrologFile(args.model)
    examples = list(read_examples(*args.examples))
    if len(examples) == 0:
        logging.getLogger('problog_lfi').warn('no examples specified')
    else:
        logging.getLogger('problog_lfi').info('Number of examples: %s' % len(examples))
    options = vars(args)
    del options['examples']

    try:
        results = run_lfi(program, examples, knowledge=knowledge, **options)

        for n in results[2]:
            n.loc = program.lineno(n.location)
        retcode = result_handler((True, results), output=outf)
    except Exception as err:
        trace = traceback.format_exc()
        err.trace = trace
        retcode = result_handler((False, err), output=outf)

    if args.output is not None:
        outf.close()

    if retcode:
        sys.exit(retcode)


def print_result(d, output, precision=8):
    success, d = d
    if success:
        score, weights, names, iterations, lfi = d
        weights = list(map(lambda x: round(x, precision), weights))
        print (score, weights, names, iterations, file=output)
        return 0
    else:
        print (process_error(d), file=output)
        return 1


def print_result_json(d, output, precision=8):
    import json
    success, d = d
    if success:
        score, weights, names, iterations, lfi = d
        results = {'SUCCESS': True,
                   'score': score,
                   'iterations': iterations,
                   'weights': [[str(n), round(w, precision), n.loc[1], n.loc[2]]
                               for n, w in zip(names, weights)],
                   'model': lfi.get_model()
                   }
        print (json.dumps(results), file=output)
    else:
        results = {'SUCCESS': False, 'err': vars(d)}
        print (json.dumps(results), file=output)
    return 0


if __name__ == '__main__':
    main(sys.argv[1:])



#! /usr/bin/env python

from __future__ import print_function
import sys
import os


version_file = os.path.join(os.path.abspath(os.path.dirname(__file__)), 'problog/version.py')

version = {}
with open(version_file) as fp:
    exec(fp.read(), version)
version = version['version']

if __name__ == '__main__' and len(sys.argv) == 1:
    from problog import setup as problog_setup
    problog_setup.install()
elif __name__ == '__main__':

    from setuptools import setup, find_packages
    from setuptools.command.install import install


    class ProbLogInstall(install):
        def run(self):
            install.run(self)
            before_dir = os.getcwd()
            sys.path.insert(0, self.install_lib)
            from problog import setup as problog_setup
            try:
                problog_setup.install()
            except Exception as err:
                print ('Optional ProbLog installation failed: %s' % err, file=sys.stderr)
            os.chdir(before_dir)

    package_data = {
        'problog': [
            'bin/darwin/dsharp', 
            'bin/darwin/maxsatz', 
            'bin/linux/dsharp', 
            'bin/linux/maxsatz',
            'bin/source/maxsatz/maxsatz2009.c',
            'bin/windows/dsharp.exe',
            'bin/windows/maxsatz.exe',
            'bin/windows/libgcc_s_dw2-1.dll',
            'bin/windows/libstdc++-6.dll',
            'lib/sdd/*.h',
            'lib/sdd/*.c',
            'lib/sdd/linux/libsdd.so',
            'lib/sdd/darwin/libsdd.a',
            'lib/sdd/*.py',
            'web/*.py',
            'web/editor_local.html'
            'web/editor_adv.html',
            'web/js/problog_editor.js',
            'library/lists.pl',
            'library/apply.pl',
            'library/control.pl'
        ]
    }


    setup(
        name='problog',
        version=version,
        description='ProbLog2: Probabilistic Logic Programming toolbox',
        url='https://dtai.cs.kuleuven.be/problog',
        author='ProbLog team',
        author_email='anton.dries@cs.kuleuven.be',
        license='Apache Software License',
        classifiers=[
            'Development Status :: 4 - Beta',
            'License :: OSI Approved :: Apache Software License',
            'Intended Audience :: Science/Research',
            'Programming Language :: Python :: 2',
            'Programming Language :: Python :: 2.7',
            'Programming Language :: Python :: 3',
            'Programming Language :: Python :: 3.3',
            'Programming Language :: Python :: 3.4',
            'Programming Language :: Prolog',
            'Topic :: Scientific/Engineering :: Artificial Intelligence',
        ],
        keywords='prolog probabilistic logic',
        packages=find_packages(),
        entry_points={
            'console_scripts': ['problog=problog.tasks:main']       
        },
        package_data=package_data,
        cmdclass={
            'install': ProbLogInstall
        }
    )


def increment_release(v):
    v = v.split('.')
    if len(v) == 4:
        v = v[:3] + [str(int(v[3]) + 1)]
    else:
        v = v[:4]
    return '.'.join(v)


def increment_dev(v):
    v = v.split('.')
    if len(v) == 4:
        v = v[:3] + [str(int(v[3]) + 1), 'dev1']
    else:
        v = v[:4] + ['dev' + str(int(v[4][3:]) + 1)]
    return '.'.join(v)


def increment_version_dev():
    v = increment_dev(version)
    os.path.dirname(__file__)
    with open(version_file, 'w') as f:
        f.write("version = '%s'\n" % v)


def increment_version_release():
    v = increment_release(version)
    with open(version_file, 'w') as f:
        f.write("version = '%s'\n" % v)


"""
Part of the ProbLog distribution.

Copyright 2015 KU Leuven, DTAI Research Group

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
import unittest

import glob, os, traceback, sys

if __name__ == '__main__' :
    sys.path.insert(0,os.path.abspath(os.path.join(os.path.dirname(__file__), '../../')))

from problog import root_path

from problog.setup import install
from problog.program import PrologFile, DefaultPrologParser, ExtendedPrologFactory
from problog.nnf_formula import NNF
from problog.sdd_formula import SDD
from problog.evaluator import SemiringProbability, SemiringLogProbability

class TestDummy(unittest.TestCase):

    def test_dummy(self) : pass

class TestSystemSDD(unittest.TestCase) :

    def setUp(self) :

        try :
            self.assertSequenceEqual = self.assertItemsEqual
        except AttributeError :
            self.assertSequenceEqual = self.assertCountEqual

class TestSystemNNF(unittest.TestCase) :

    def setUp(self) :

        try :
            self.assertSequenceEqual = self.assertItemsEqual
        except AttributeError :
            self.assertSequenceEqual = self.assertCountEqual



def read_result(filename) :
    results = {}
    with open( filename ) as f :
        reading = False
        for l in f :
            l = l.strip()
            if l.startswith('%Expected outcome:') :
                reading = True
            elif reading :
                if l.lower().startswith('% error') :
                    return l[len('% error'):].strip()
                elif l.startswith('% ') :
                    query, prob = l[2:].rsplit(None,1)
                    results[query.strip()] = float(prob.strip())
                else :
                    reading = False
            if l.startswith('query(') and l.find('% outcome:') >= 0 :
                pos = l.find('% outcome:')
                query = l[6:pos].strip().rstrip('.').rstrip()[:-1]
                prob = l[pos+10:]
                results[query.strip()] = float(prob.strip())
    return results

def createSystemTestSDD(filename, logspace=False) :

    correct = read_result(filename)

    def test(self) :
        try :
            parser = DefaultPrologParser(ExtendedPrologFactory())
            sdd = SDD.createFrom(PrologFile(filename, parser=parser))

            if logspace :
                semiring = SemiringLogProbability()
            else :
                semiring = SemiringProbability()

            computed = sdd.evaluate(semiring=semiring)
            computed = { str(k) : v for k,v in computed.items() }
        except Exception as err :
            e = err
            computed = None

        if computed is None :
            self.assertEqual(correct, type(e).__name__)
        else :
            self.assertIsInstance( correct, dict )
            self.assertSequenceEqual(correct, computed)

            for query in correct :
                self.assertAlmostEqual(correct[query], computed[query], msg=query)

    return test


def createSystemTestNNF(filename, logspace=False) :

    correct = read_result(filename)

    def test(self) :
        try :
            parser = DefaultPrologParser(ExtendedPrologFactory())
            sdd = NNF.createFrom(PrologFile(filename, parser=parser))

            if logspace :
                semiring = SemiringLogProbability()
            else :
                semiring = SemiringProbability()

            computed = sdd.evaluate(semiring=semiring)
            computed = { str(k) : v for k,v in computed.items() }
        except Exception as err :
            e = err
            computed = None

        if computed is None :
            self.assertEqual(correct, type(e).__name__)
        else :
            self.assertIsInstance( correct, dict )
            self.assertSequenceEqual(correct, computed)

            for query in correct :
                self.assertAlmostEqual(correct[query], computed[query], msg=query)

    return test


if __name__ == '__main__' :
    filenames = sys.argv[1:]
else :
    filenames = glob.glob( root_path('test', '*.pl' ) )

for testfile in filenames :
    # testname = 'test_system_' + os.path.splitext(os.path.basename(testfile))[0]
    # setattr( TestSystemSDD, testname, createSystemTestSDD(testfile) )
    # setattr( TestSystemNNF, testname, createSystemTestNNF(testfile) )

    testname = 'test_system_' + os.path.splitext(os.path.basename(testfile))[0]
    setattr( TestSystemSDD, testname, createSystemTestSDD(testfile, True) )
    setattr( TestSystemNNF, testname, createSystemTestNNF(testfile, True) )


if __name__ == '__main__' :
    suite = unittest.TestLoader().loadTestsFromTestCase(TestSystemSDD)
    unittest.TextTestRunner(verbosity=2).run(suite)
    suite = unittest.TestLoader().loadTestsFromTestCase(TestSystemNNF)
    unittest.TextTestRunner(verbosity=2).run(suite)



"""
problog.program - Representation of Logic Programs
--------------------------------------------------

Provides tools for loading logic programs.

..
    Part of the ProbLog distribution.

    Copyright 2015 KU Leuven, DTAI Research Group

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""
from __future__ import print_function

from .errors import GroundingError
from .logic import Term, Var, Constant, AnnotatedDisjunction, Clause, And, Or, Not

from .parser import DefaultPrologParser, Factory
from .core import ProbLogError

import os
import sys


class LogicProgram(object):
    """LogicProgram"""

    def __init__(self, source_root='.', source_files=None, line_info=None, **extra_info):
        if source_files is None:
            source_files = [None]
        if line_info is None:
            line_info = [None]
        self.source_root = source_root
        self.source_files = source_files
        self.source_parent = [None]
        self.extra_info = extra_info
        # line_info should be array, corresponding to 'source_files'.
        self.line_info = line_info

    def __iter__(self):
        """Iterator for the clauses in the program."""
        raise NotImplementedError("LogicProgram.__iter__ is an abstract method.")

    def add_clause(self, clause):
        """Add a clause to the logic program.

        :param clause: add a clause
        """
        raise NotImplementedError("LogicProgram.addClause is an abstract method.")

    def add_fact(self, fact):
        """Add a fact to the logic program.

        :param fact: add a fact
        """
        raise NotImplementedError("LogicProgram.addFact is an abstract method.")

    def __iadd__(self, clausefact):
        """Add clause or fact using the ``+=`` operator."""
        if isinstance(clausefact, Or):
            heads = clausefact.to_list()
            # TODO move this to parser code
            for head in heads:
                if not type(head) == Term:
                    # TODO compute correct location
                    raise GroundingError("Unexpected fact '%s'" % head)
                elif len(heads) > 1 and head.probability is None:
                    raise GroundingError("Non-probabilistic head in multi-head clause '%s'" % head)
            self.add_clause(AnnotatedDisjunction(heads, Term('true')))
        elif isinstance(clausefact, AnnotatedDisjunction):
            self.add_clause(clausefact)
        elif isinstance(clausefact, Clause):
            self.add_clause(clausefact)
        elif type(clausefact) == Term:
            self.add_fact(clausefact)
        else:
            raise GroundingError("Unexpected fact '%s'" % clausefact,
                                 self.lineno(clausefact.location))
        return self

    @classmethod
    def create_from(cls, src, force_copy=False, **extra):
        """Create a LogicProgram of the current class from another LogicProgram.

        :param src: logic program to convert
        :type src: :class:`.LogicProgram`
        :param force_copy: default False, If true, always create a copy of the original logic \
        program.
        :type force_copy: bool
        :param extra: additional arguments passed to all constructors and action functions
        :returns: LogicProgram that is (externally) identical to given one
        :rtype: object of the class on which this method is invoked

        If the original LogicProgram already has the right class and force_copy is False, then \
        the original program is returned.
        """
        return cls.createFrom(src, force_copy=force_copy, **extra)

    # noinspection PyPep8Naming
    @classmethod
    def createFrom(cls, src, force_copy=False, **extra):
        """Create a LogicProgram of the current class from another LogicProgram.

        :param src: logic program to convert
        :type src: :class:`.LogicProgram`
        :param force_copy: default False, If true, always create a copy of the original logic \
        program.
        :type force_copy: bool
        :param extra: additional arguments passed to all constructors and action functions
        :returns: LogicProgram that is (externally) identical to given one
        :rtype: object of the class on which this method is invoked

        If the original LogicProgram already has the right class and force_copy is False, then \
        the original program is returned.
        """
        if not force_copy and src.__class__ == cls:
            return src
        else:
            obj = cls(**extra)
            if hasattr(src, 'extra_info'):
                obj.extra_info.update(src.extra_info)
            if hasattr(src, 'source_root') and hasattr(src, 'source_files'):
                obj.source_root = src.source_root
                obj.source_files = src.source_files[:]
                obj.source_parent = src.source_parent[:]
            if hasattr(src, 'line_info'):
                obj.line_info = src.line_info[:]
            for clause in src:
                obj += clause
            return obj

    def lineno(self, char, force_filename=False):
        """Transform character position to line:column format.

        :param char: character position
        :param force_filename: always add filename even for top-level file
        :return: line, column (or None if information is not available)
        """
        # Input should be tuple (file, char)
        if isinstance(char, tuple):
            fn, char = char
        else:
            fn = 0

        if self.line_info[fn] is None or char is None:
            # No line info available
            return None
        else:
            import bisect
            i = bisect.bisect_right(self.line_info[fn], char)
            lineno = i
            charno = char - self.line_info[fn][i - 1]
            if fn == 0 and not force_filename:
                filename = None
            else:
                filename = self.source_files[fn]
            return filename, lineno, charno


class SimpleProgram(LogicProgram):
    """LogicProgram implementation as a list of clauses."""

    def __init__(self):
        LogicProgram.__init__(self)
        self.__clauses = []

    def add_clause(self, clause):
        """Add a clause to the logic program.

        :param clause: add a clause
        """
        if type(clause) is list:
            for c in clause:
                self.__clauses.append(c)
        else:
            self.__clauses.append(clause)

    def add_fact(self, fact):
        """Add a fact to the logic program.

        :param fact: add a fact
        """
        self.__clauses.append(fact)

    def __iter__(self):
        return iter(self.__clauses)


class PrologString(LogicProgram):
    """Read a logic program from a string of ProbLog code."""

    def __init__(self, string, parser=None, factory=None, source_root='.', source_files=None,
                 identifier=0):
        self.__string = string
        self.__program = None
        self.__identifier = identifier
        lines = [self._find_lines(string)]
        if parser is None:
            if factory is None:
                factory = DefaultPrologFactory(identifier=identifier)
            self.parser = DefaultPrologParser(factory)
        else:
            self.parser = parser

        LogicProgram.__init__(self, source_root=source_root, source_files=source_files,
                              line_info=lines, factory=factory, parser=parser)

    def _program(self):
        """Parsed program"""
        if self.__program is None:
            self.__program = self.parser.parseString(self.__string)
        return self.__program

    def _find_lines(self, s):
        """Find line-end positions."""
        lines = [-1]
        f = s.find('\n')
        while f >= 0:
            lines.append(f)
            f = s.find('\n', f + 1)
        lines.append(len(s))
        return lines

    def __iter__(self):
        """Iterator for the clauses in the program."""
        return iter(self._program())

    def __getitem__(self, sl):
        program = self._program()
        return program[sl]

    def add_clause(self, clause):
        """Add a clause to the logic program.

        :param clause: add a clause
        """
        raise AttributeError('not supported')

    def add_fact(self, fact):
        """Add a fact to the logic program.

        :param fact: add a fact
        """
        raise AttributeError('not supported')


class PrologFile(PrologString):
    """LogicProgram implementation as a pointer to a Prolog file.

    :param filename: filename of the Prolog file (optional)
    :param identifier: index of the file (in case of multiple files)
    :type filename: string
    """

    def __init__(self, filename, parser=None, factory=None, identifier=0):
        if filename == '-':
            source_root = ''
            source_files = ['-']
            source_text = sys.stdin.read()
        else:
            rootfile = os.path.abspath(filename)
            source_root = os.path.dirname(filename)
            source_files = [rootfile]
            try:
                with open(filename) as f:
                    source_text = f.read()
            except IOError as err:
                raise ProbLogError(str(err))
        PrologString.__init__(self, source_text, parser=parser, factory=factory,
                              source_root=source_root, source_files=source_files,
                              identifier=identifier)

    def add_clause(self, clause):
        """Add a clause to the logic program.

        :param clause: add a clause
        """
        raise AttributeError('not supported')

    def add_fact(self, fact):
        """Add a fact to the logic program.

        :param fact: add a fact
        """
        raise AttributeError('not supported')


class PrologFactory(Factory):
    """Factory object for creating suitable objects from the parse tree."""

    def __init__(self, identifier=0):
        self.loc_id = identifier

    def build_program(self, clauses):
        return clauses

    def build_function(self, functor, arguments, location=None, **extra):
        return Term(functor, *arguments, location=(self.loc_id, location), **extra)

    def build_variable(self, name, location=None):
        return Var(name, location=(self.loc_id, location))

    def build_constant(self, value, location=None):
        return Constant(value, location=(self.loc_id, location))

    def build_binop(self, functor, operand1, operand2, function=None, location=None, **extra):
        return self.build_function("'" + functor + "'", (operand1, operand2), location=location, **extra)

    def build_directive(self, functor, operand, **extra):
        head = self.build_function('_directive', [])
        return self.build_clause(functor, [head], operand, **extra)

    def build_unop(self, functor, operand, location=None, **extra):
        if functor == '-' and operand.is_constant() and \
                (operand.is_float() or operand.is_integer()):
            return Constant(-operand.value)
        return self.build_function("'" + functor + "'", (operand,), location=location, **extra)

    def build_list(self, values, tail=None, location=None, **extra):
        if tail is None:
            current = Term('[]')
        else:
            current = tail
        for value in reversed(values):
            current = self.build_function('.', (value, current), location=location)
        return current

    def build_string(self, value, location=None):
        return self.build_constant('"' + value + '"', location=location)

    def build_cut(self, location=None):
        raise AttributeError('not supported')

    # noinspection PyUnusedLocal
    def build_clause(self, functor, operand1, operand2, location=None, **extra):
        heads = operand1
        # TODO move this to parser code
        for head in heads:
            if not type(head) == Term:
                # TODO compute correct location
                raise GroundingError("Unexpected clause head '%s'" % head)
            elif len(heads) > 1 and head.probability is None:
                raise GroundingError("Non-probabilistic head in multi-head clause '%s'" % head)
        if len(heads) > 1:
            return AnnotatedDisjunction(heads, operand2, location=(self.loc_id, location))
        else:
            return Clause(operand1[0], operand2, location=(self.loc_id, location))

    # noinspection PyUnusedLocal
    def build_disjunction(self, functor, operand1, operand2, location=None, **extra):
        return Or(operand1, operand2, location=(self.loc_id, location))

    # noinspection PyUnusedLocal
    def build_conjunction(self, functor, operand1, operand2, location=None, **extra):
        return And(operand1, operand2, location=(self.loc_id, location))

    # noinspection PyUnusedLocal
    def build_not(self, functor, operand, location=None, **extra):
        return Not(functor, operand, location=(self.loc_id, location))

    def build_probabilistic(self, operand1, operand2, location=None, **extra):
        operand2.probability = operand1
        return operand2

    def _uncurry(self, term, func=None):
        if func is None:
            func = term.functor

        body = []
        current = term
        while isinstance(current, Term) and current.functor == func:
            body.append(current.args[0])
            current = current.args[1]
        body.append(current)
        return body


class ExtendedPrologFactory(PrologFactory):
    """Prolog with some extra syntactic sugar.

    Non-standard syntax:
    - Negative head literals [Meert and Vennekens, PGM 2014]: 0.5::\+a :- b.
    """

    def __init__(self, identifier=0):
        PrologFactory.__init__(self, identifier)
        self.neg_head_lits = dict()

    def _update_functors(self, t):
        """Adapt functors that appear as a negative literal to be f_p and f_n
        where f appears in the head.
        TODO: Should be implemented using a more general visitor pattern
        """
        if type(t) is Clause:
            self._update_functors(t.head)
        elif type(t) is AnnotatedDisjunction:
            self._update_functors(t.heads)
        elif type(t) is Term:
            if t.signature in self.neg_head_lits:
                t.functor = self.neg_head_lits[t.signature]['p']
        elif type(t) is Not:
            self._update_functors(t.child)
        elif type(t) is Or or type(t) is And:
            self._update_functors(t.op1)
            self._update_functors(t.op2)
        elif type(t) is None or type(t) is Var or type(t) is Constant:
            pass
        elif type(t) is list:
            for term in t:
                self._update_functors(term)
        else:
            raise Exception("Unknown type: {} -- {}".format(t, type(t)))

    def build_program(self, clauses):
        """Update functor f that appear as a negative head literal to f_p and
        :param clauses:
        :return:
        """
        # f_n
        for clause in clauses:
            self._update_functors(clause)

        # Add extra rule for a functor f that appears a as a negative head
        # literal such that:
        # f :- f_p, \+f_n.
        for k, v in self.neg_head_lits.items():
            cur_vars = [Var("V{}".format(i)) for i in range(v['c'])]
            new_clause = Clause(Term(v['f'], *cur_vars),
                                And(Term(v['p'], *cur_vars), Not('\+', Term(v['n'], *cur_vars))))
            clauses.append(new_clause)
        return clauses

    def neg_head_literal_to_pos_literal(self, literal):
        """Translate a negated literal into a positive literal and remember
        the literal to update the complete program later (in build_program).
        :param literal:
        :return:
        """
        literal = abs(literal)
        if literal.signature not in self.neg_head_lits:
            self.neg_head_lits[literal.signature] = {
                'c': literal.arity,
                'p': literal.functor + "_p",
                'n': literal.functor + "_n",
                'f': literal.functor
            }
        literal.functor = self.neg_head_lits[literal.signature]['n']
        return literal

    def build_probabilistic(self, operand1, operand2, location=None, **extra):
        """Detect probabilistic negated head literal and translate to positive literal
        :param operand1:
        :param operand2:
        :param location:
        :param extra:
        :return:
        """
        if ('unaryop' in extra and extra['unaryop'] == '\\+') or operand2.is_negated():
            operand2 = self.neg_head_literal_to_pos_literal(operand2)
        operand2.probability = operand1
        return operand2

    def build_clause(self, functor, operand1, operand2, location=None, **extra):
        """Detect deterministic head literal and translate to positive literal
        :param functor:
        :param operand1:
        :param operand2:
        :param location:
        :param extra:
        :return:
        """
        heads = operand1
        new_heads = []
        for head in heads:
            if type(head) == Not:
                new_head = self.neg_head_literal_to_pos_literal(head)
                new_heads.append(new_head)
            else:
                new_heads.append(head)
        return super(ExtendedPrologFactory, self).build_clause(functor, new_heads, operand2, location, **extra)

DefaultPrologFactory = ExtendedPrologFactory
